"""Instrument driver for the Zürich Instruments HDAWG."""

import enum
import logging
import json
import re
import time
import typing
from pathlib import Path
from typing import Optional, Union, List, Dict, Tuple, Any

import jsonschema  # type: ignore
import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method


# Lazy import of the zhinst module. See the function _import_modules() below.
if typing.TYPE_CHECKING:
    import zhinst.ziPython
    import zhinst.utils
else:
    zhinst = None


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


def _import_modules() -> None:
    """Import the zhinst library.

    This import is done in a function, instead of at the top-level,
    to avoid an unnecessary dependency for programs that do not access
    the instrument directly.
    """
    global zhinst
    if zhinst is None:
        import zhinst.ziPython  # pylint: disable=W0621
        import zhinst.utils


# Compile and upload timeout in seconds.
_COMPILE_TIMEOUT = 30
_UPLOAD_TIMEOUT = 30

# Location of the default command table schema
_THIS_MODULE_PATH = Path(__file__).resolve()
_DEFAULT_CT_TABLE_SCHEMA_FILENAME = "hdawg_command_table.schema"
_DEFAULT_CT_TABLE_SCHEMA_PATH = Path(_THIS_MODULE_PATH.parent, _DEFAULT_CT_TABLE_SCHEMA_FILENAME)

# Sequencer code replacement variables must start with a literal $, followed by at least one letter followed by zero or
# more alphanumeric characters or underscores.
SEQC_PAR_PATTERN = re.compile(r"\$[A-Za-z][A-Za-z0-9_]*", re.ASCII)


# Status enumerations.
class CompilerStatus(enum.IntEnum):
    """Enumeration of compiler process status."""
    NOT_READY = -1
    READY = 0
    READY_WITH_ERRORS = 1
    READY_WITH_WARNINGS = 2


class UploadStatus(enum.IntEnum):
    """Enumeration of ELF upload process status."""
    WAITING = -1
    DONE = 0
    FAILED = 1
    BUSY = 2


class ZurichInstruments_HDAWG(QMI_Instrument):
    """Instrument driver for the Zürich Instruments HDAWG."""

    NUM_AWGS = 4
    NUM_CHANNELS = 8
    NUM_CHANNELS_PER_AWG = 2

    def __init__(self, context: QMI_Context, name: str, server_host: str, server_port: int, device_name: str) -> None:
        """Initialize driver.

        We connect to a specific HDAWG via a Data Server, which is a process running on some computer.

        Parameters:
            name: Name for this instrument instance.
            server_host: Host where the ziDataServer process is running.
            server_port: TCP port there the ziDataServer process can be reached.
            device_name: Name of the HDAWG device (typically "devNNNN").
        """
        super().__init__(context, name)

        self._server_host = server_host
        self._server_port = server_port
        self._device_name = device_name

        self._daq_server: Optional[zhinst.ziPython.ziDAQServer] = None
        self._awg_module: Optional[zhinst.ziPython.AwgModule] = None

        self._last_compilation_successful = False

        # Import the "zhinst" module.
        _import_modules()

    @rpc_method
    def open(self) -> None:

        self._check_is_closed()
        _logger.info("[%s] Opening connection to instrument", self._name)

        # This may fail!
        self._daq_server = zhinst.ziPython.ziDAQServer(self._server_host, self._server_port)

        # Connect to the device via Ethernet.
        # If the device is already connected, this is a no-op.
        assert self._daq_server is not None
        self._daq_server.connectDevice(self._device_name, "1GbE")

        self._awg_module = self._daq_server.awgModule()

        assert self._awg_module is not None
        self._awg_module.set("awgModule/device", self._device_name)
        self._awg_module.set("awgModule/index", 0)  # only support 1x8 mode, so only one AWG module

        # Verify that the AWG thread is not running.
        assert self._awg_module.finished()

        # Start the AWG thread.
        self._awg_module.execute()

        # Verify that the AWG thread is running.
        assert not self._awg_module.finished()

        super().open()

    @rpc_method
    def close(self) -> None:

        self._check_is_open()
        assert self._awg_module is not None

        _logger.info("[%s] Closing connection to instrument", self._name)
        # Verify that the AWG thread is running.
        assert not self._awg_module.finished()

        # Stop the AWG module thread.
        self._awg_module.finish()

        # Verify that the AWG thread is no longer running.
        assert self._awg_module.finished()

        self._awg_module = None

        assert self._daq_server is not None
        self._daq_server.disconnect()
        self._daq_server = None

        super().close()

    def _set_dev_value(self, node_path: str, value: Union[int, float, str]) -> None:
        """Set an arbitrary value in the the device node tree.

        Parameters:
            node_path:  Path in the device tree, relative to the "/devNNNN/" subtree.
            value:      Value to write.
        """
        assert self._daq_server is not None
        self._daq_server.set('/' + self._device_name + '/' + node_path, value)

    def _set_dev_int(self, node_path: str, value: int) -> None:
        """Set an integer value in the device node tree.

        Parameters:
            node_path: Path in the device tree, relative to the "/devNNNN/" subtree.
            value: Integer value to write.
        """
        assert self._daq_server is not None
        self._daq_server.setInt('/' + self._device_name + '/' + node_path, value)

    def _set_dev_double(self, node_path: str, value: float) -> None:
        """Set a floating point value in the device node tree.

        Parameters:
            node_path: Path in the device tree, relative to the "/devNNNN/" subtree.
            value: Floating point value to write.
        """
        assert self._daq_server is not None
        self._daq_server.setDouble('/' + self._device_name + '/' + node_path, value)

    def _get_dev_int(self, node_path: str) -> int:
        """Return an integer value from the device node tree.

        Parameters:
            node_path: Path in the device tree, relative to the "/devNNNN/" subtree.
        """
        assert self._daq_server is not None
        return self._daq_server.getInt('/' + self._device_name + '/' + node_path)

    def _get_dev_double(self, node_path: str) -> float:
        """Return a floating point value from the device node tree.

        Parameters:
            node_path: Path in the device tree, relative to the "/devNNNN/" subtree.
        """
        assert self._daq_server is not None
        return self._daq_server.getDouble('/' + self._device_name + '/' + node_path)

    @staticmethod
    def _process_parameter_replacements(sequencer_program: str, replacements: Dict[str, Union[str, int, float]]) -> str:
        """Process parameter replacements for sequencer code."""
        for parameter, replacement in replacements.items():
            # Convert replacements to "str".
            if isinstance(replacement, (int, float)):
                replacement = repr(replacement)

            # At this point, the replacement value should be a string.
            if not isinstance(replacement, str):
                raise ValueError("Cannot handle replacement value of type {!r}".format(type(replacement)))

            # Perform the replacement.
            if SEQC_PAR_PATTERN.fullmatch(parameter):
                # Replace parameter, respecting word boundaries only (note: '$' is already a word boundary)
                parameter_pattern = f"\\{parameter}\\b"  # escape the '$' and add word boundary match
                sequencer_program = re.sub(parameter_pattern, replacement, sequencer_program)
            else:
                raise NameError("Replacement parameter has an invalid name: {}".format(parameter))

        # Check if there are any unreplaced parameters left in the source code; this will not compile.
        leftover_parameters = SEQC_PAR_PATTERN.findall(sequencer_program)
        if leftover_parameters:
            raise KeyError("Variables left in sequencer program that were not in replacement dictionary: {}".format(
                ', '.join(leftover_parameters)
            ))

        return sequencer_program

    @staticmethod
    def _check_program_not_empty(sequencer_program: str):
        """Check if the program is non-empty (compiler silently ignores empty programs)."""
        # Filter out lines that start with // or /* (comments) or are empty.
        seqc_statements = list(filter(
            lambda s: not (s.startswith("//") or s.startswith("/*") or len(s) == 0),
            [s.strip() for s in sequencer_program.split('\n')]
        ))

        # Check if there are any lines left (we do not check if that is executable code; the compiler will do that).
        if len(seqc_statements) == 0:
            raise ValueError("Source string does not contain executable statements")

    def _wait_compile(self, sequencer_program: str) -> CompilerStatus:
        """Start a compilation of a sequencer program."""

        self._check_is_open()
        assert self._awg_module is not None

        # Compile the sequencer program with replacements made.
        self._awg_module.set("awgModule/compiler/sourcestring", sequencer_program)

        # Poll the AWG module to check if compilation progress.
        # See https://docs.zhinst.com/labone_programming_manual/awg_module.html#_awg_module_parameters for how to query
        # the progress.
        compilation_start_time = time.monotonic()
        _logger.debug("Compilation started ... ")
        compilation_status = CompilerStatus(self._awg_module.getInt("awgModule/compiler/status"))
        while compilation_status == CompilerStatus.NOT_READY:
            time.sleep(0.1)
            compilation_status = CompilerStatus(self._awg_module.getInt("awgModule/compiler/status"))
            if time.monotonic() - compilation_start_time > _COMPILE_TIMEOUT:
                raise RuntimeError("Compilation process timed out (timeout={})".format(_COMPILE_TIMEOUT))
        compilation_end_time = time.monotonic()
        _logger.debug(
            "Compilation finished in %.1f seconds (status=%d)",
            compilation_end_time - compilation_start_time,
            compilation_status
        )

        return compilation_status

    def _wait_upload(self) -> UploadStatus:
        """Poll ELF upload progress and check result."""

        self._check_is_open()
        assert self._awg_module is not None

        # Poll the AWG module to check ELF upload progress.
        upload_start_time = time.monotonic()
        _logger.debug("Polling ELF upload status ...")
        upload_progress = self._awg_module.getDouble("awgModule/progress")
        upload_status = UploadStatus(self._awg_module.getInt("awgModule/elf/status"))
        while upload_progress < 1.0 and upload_status in (UploadStatus.WAITING, UploadStatus.DONE, UploadStatus.BUSY):
            time.sleep(0.1)
            upload_progress = self._awg_module.getDouble("awgModule/progress")
            upload_status = UploadStatus(self._awg_module.getInt("awgModule/elf/status"))
            if time.monotonic() - upload_start_time > _UPLOAD_TIMEOUT:
                raise RuntimeError("Upload process timed out (timeout={})".format(_UPLOAD_TIMEOUT))
        upload_end_time = time.monotonic()
        _logger.debug(
            "ELF upload finished in %.1f seconds (status=%d)",
            upload_end_time - upload_start_time,
            upload_status
        )

        return upload_status

    def _interpret_compilation_result_is_ok(self, compilation_result: CompilerStatus) -> bool:
        """Interpret compilation result."""

        self._check_is_open()
        assert self._awg_module is not None

        if compilation_result == CompilerStatus.READY:
            # Successful compilation; proceed.
            ok_to_proceed = True
        elif compilation_result == CompilerStatus.READY_WITH_ERRORS:
            # Compilation finished with errors.
            error_message = self._awg_module.getString("awgModule/compiler/statusstring")
            _logger.error("Compilation finished with errors: %s", error_message)
            ok_to_proceed = False
        elif compilation_result == CompilerStatus.READY_WITH_WARNINGS:
            # Compilation finished with warnings.
            warning_message = self._awg_module.getString("awgModule/compiler/statusstring")
            _logger.warning("Compilation finished with warnings: %s", warning_message)
            ok_to_proceed = False
        else:
            raise ValueError("Unknown compiler status: {}".format(compilation_result))

        return ok_to_proceed

    def _interpret_upload_result_is_ok(self, upload_result: UploadStatus) -> bool:
        """Interpret upload result."""

        self._check_is_open()
        assert self._awg_module is not None

        if upload_result == UploadStatus.DONE:
            # Successful upload; proceed.
            ok_to_proceed = True
        elif upload_result == UploadStatus.FAILED:
            # Upload failed.
            _logger.error("ELF upload failed")
            ok_to_proceed = False
        elif upload_result == UploadStatus.BUSY:
            # Upload still in progress; we should never get here.
            _logger.error("ELF upload in progress but aborted for unknown reason.")
            ok_to_proceed = False
        else:
            raise ValueError("Unknown upload status: {}".format(upload_result))

        return ok_to_proceed

    @rpc_method
    def set_node_value(self, node_path: str, value: Union[int, float, str]) -> None:
        """Write an arbitrary value to the device node tree.

        Requires LabOne version 21.08 or newer.

        Parameters:
            node_path:  Path in the device tree, relative to the "/devNNNN/" subtree.
            value:      Value to write.
        """
        self._set_dev_value(node_path, value)

    @rpc_method
    def get_node_int(self, node_path: str) -> int:
        """Get an integer value in the device node tree.

        Parameters:
            node_path:  Path in the device tree, relative to the "/devNNNN/" subtree.
        """
        return self._get_dev_int(node_path)

    @rpc_method
    def set_node_int(self, node_path: str, value: int) -> None:
        """Set an integer value in the device node tree.

        Parameters:
            node_path: Path in the device tree, relative to the "/devNNNN/" subtree.
            value: Integer value to write.
        """
        self._set_dev_int(node_path, value)

    @rpc_method
    def get_node_double(self, node_path: str) -> float:
        """Get a floating point value in the device node tree.

        Parameters:
            node_path:  Path in the device tree, relative to the "/devNNNN/" subtree.
        """
        return self._get_dev_double(node_path)

    @rpc_method
    def set_node_double(self, node_path: str, value: float) -> None:
        """Set a floating point value in the device node tree.

        Parameters:
            node_path: Path in the device tree, relative to the "/devNNNN/" subtree.
            value: Floating point value to write.
        """
        self._set_dev_double(node_path, value)

    @rpc_method
    def compile_and_upload(self,
                           sequencer_program: str,
                           replacements: Optional[Dict[str, Union[str, int, float]]] = None
                           ) -> None:
        """Compile and upload the sequencer_program, after performing textual replacements.

        This function combines compilation followed by upload to the AWG if compilation was successful. This is forced
        by the HDAWG API.

        Notes on parameter replacements:
         - Parameters must adhere to the following format: $[A-Za-z][A-Za-z0-9]+ (literal $, followed by at least one
            letter followed by zero or more alphanumeric characters or underscores). Both the key in the replacements
            dictionary and the parameter reference in the sequencer code must adhere to this format.
         - Replacement respects word boundaries. The inclusion of the '$' allows to concatenate values in the sequencer
            program code, e.g. "wave w = "$TYPE$LENGTH;" to achieve "wave w = "sin1024";", but be careful that you
            don't accidentally create new parameters by concatenation.
         - Replacement values must be of type str, int or float.

        Parameters:
            sequencer_program: Full text of the sequencer script.
            replacements: Optional dictionary of (parameter, value) pairs. Every occurrence of the parameter in the
                          sequencer program will be replaced  with the specified value.
        """

        self._check_is_open()
        assert self._awg_module is not None

        # Perform parameter replacements.
        if replacements is None:
            replacements = {}
        sequencer_program = self._process_parameter_replacements(sequencer_program, replacements)
        self._check_program_not_empty(sequencer_program)

        # Compile.
        compilation_result = self._wait_compile(sequencer_program)
        result_ok = self._interpret_compilation_result_is_ok(compilation_result)
        if result_ok:
            # Allow some time for upload to start.
            time.sleep(1.0)

            # Wait for upload to finish.
            upload_result = self._wait_upload()
            result_ok = self._interpret_upload_result_is_ok(upload_result)

        # Done; store result.
        self._last_compilation_successful = result_ok

    @rpc_method
    def compilation_successful(self) -> bool:
        """Query result of compilation process (sequence compilation and ELF upload).

        Note: this will only return a meaningful value after starting at least one compilation process. The method
        returns False if there was no previous compilation result.
        """
        return self._last_compilation_successful

    @rpc_method
    def upload_waveform(self,
                        awg_index: int,
                        waveform_index: int,
                        wave1: np.ndarray,
                        wave2: Optional[np.ndarray] = None,
                        markers: Optional[np.ndarray] = None
                        ) -> None:
        """Upload new waveform data to the AWG.

        The AWG must be explicitly disabled by calling `set_awg_module_enabled(0)`
        before uploading waveforms, and re-enabled by calling
        `set_awg_module_enabled(1)` after uploading waveforms.

        A waveform array inside the AWG can hold interleaved data
        for up to 2 analog waveforms and 4 marker waveforms.
        If a waveform array contains multiple waveforms, these must
        all be uploaded together.

        The AWG compiler decides how "wave" variables are mapped
        to waveform array indexes. This can be inspected in the
        "Waveform Viewer" in the LabOne user interface.

        Parameters:
            awg_index:       0-based index of the AWG (group of 2 channels).
            waveform_index:  0-based index of the waveform array.
            wave1:           Array containing floating point samples
                             in range -1.0 .. +1.0 for the first waveform.
            wave2:           Array containing floating point samples
                             in range -1.0 .. +1.0 for the second waveform.
            markers:         Array containing integer values where
                             the 4 least significant bits of each sample
                             represent the 4 marker channels.
        """
        self._check_is_open()
        assert self._daq_server is not None
        waveform_data = zhinst.utils.convert_awg_waveform(wave1, wave2, markers)
        waveform_address = f"/{self._device_name}/awgs/{awg_index}/waveform/waves/{waveform_index}"
        self._daq_server.setVector(waveform_address, waveform_data)

    @rpc_method
    def upload_waveforms(
            self,
            unpacked_waveforms: List[Tuple[int, int, np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]],
            batch_size: int = 50
    ) -> None:
        """Upload a set of new waveform data to the AWG. Works as singular waveform uploading, but creating a
        list of tuple sets of waveforms, and uploading all in one command, speeds up the upload significantly.

        Large sets of waveforms need to be batched to avoid running out of memory.

        Parameters:
            unpacked_waveforms: List of Tuples, each tuple is a collection of awg_index, waveform sequence index,
                wave1, wave2 and markers.
            batch_size: Large sets of waveforms take plenty of memory. This makes the waveforms to be sent with
                maximum sized batches. Default size is 50 waveform entries.
        """
        self._check_is_open()
        assert self._daq_server is not None

        waves_set = []
        for wf_count, sequence in enumerate(unpacked_waveforms):
            awg_index, waveform_index, wave1, wave2, markers = sequence
            wave_raw = zhinst.utils.convert_awg_waveform(wave1, wave2, markers)
            waveform_address = f'/{self._device_name}/awgs/{awg_index}/waveform/waves/{waveform_index}'
            waves_set.append((waveform_address, wave_raw))
            # Check set size against batch size
            if wf_count % batch_size == batch_size - 1:
                # send a batch and reset list
                self._daq_server.set(waves_set)
                waves_set = []

        # Send also any possible remains of the last batch
        if len(waves_set):
            self._daq_server.set(waves_set)

    @rpc_method
    def upload_command_table(self, awg_index: int, command_table_entries: List[Dict[str, Any]]) -> None:
        """Upload a new command table to the AWG.

        Command tables are needed for using e.g. `assignWaveIndex` and must be uploaded before the sequencer program
        is executed. Do not upload while a program is running. Note that each core that is used by the sequencer
        script needs a valid command table and that grouped cores need the same command table. In 1x8 mode all 4
        cores thus need identical tables.

        This routine converts the list of entries to JSON and prepends the required headers before uploading.

        The resulting table is checked against the schema before uploading. A copy of the schema is included with this
        driver and used by default; the original is available at https://docs.zhinst.com/hdawg/commandtable/v2/schema
        (note: the version on the website of Zurich instruments has a https:// URI and a # (anchor) at the end of the
        $schema attribute, which is not recognised by the jsonschema implementation for Python. The version of the
        file included here has this fixed (see also https://github.com/Julian/jsonschema/issues/569)).

        Parameters:
            awg_index:              0-based index of the AWG core to apply the table to (0 .. 3).
            command_table_entries:  actual command table as a list of entries (dicts).
        """
        self._check_is_open()
        assert self._daq_server is not None

        # Check AWG core index.
        if awg_index not in (0, 1, 2, 3):
            raise ValueError("AWG index must be in 0 .. 3")

        # Create the command table from the provided entries.
        command_table = {
            "$schema": "http://docs.zhinst.com/hdawg/commandtable/v2/schema",
            "header": {
                "version": "0.2"
            },
            "table": command_table_entries
        }

        # Load the validation schema.
        validation_schema_file = _DEFAULT_CT_TABLE_SCHEMA_PATH
        try:
            with open(validation_schema_file, 'r') as fhandle:
                validation_schema = json.load(fhandle)
        except json.JSONDecodeError as exc:
            _logger.exception("Error in decoding validation schema", exc_info=exc)
            raise ValueError("Invalid JSON") from exc

        # Validate the table against the schema.
        try:
            jsonschema.validate(command_table, schema=validation_schema)
        except jsonschema.exceptions.SchemaError as exc:
            _logger.exception("The provided schema is invalid", exc_info=exc)
            raise ValueError("Invalid schema") from exc
        except jsonschema.exceptions.ValidationError as exc:
            _logger.exception("The provided command table is not valid", exc_info=exc)
            raise ValueError("Invalid command table") from exc

        # Convert the command table to JSON and upload.
        try:
            command_table_as_json = json.dumps(command_table, allow_nan=False)
        except (TypeError, ValueError) as exc:
            raise ValueError("Invalid value in command table") from exc

        self._daq_server.setVector(
            "/{}/awgs/{}/commandtable/data".format(self._device_name, awg_index),
            command_table_as_json
        )

    @rpc_method
    def sync(self) -> None:
        """Synchronize the state of the AWG.

        This call ensures that all previous settings have taken effect on
        the instrument, and stale data is flushed from local buffers.
        """
        self._check_is_open()
        assert self._daq_server is not None
        self._daq_server.sync()

    @rpc_method
    def set_channel_grouping(self, value: int) -> None:
        """Set the channel grouping of the device.

        This QMI driver currently supports only channel grouping 2 (1x8 channels).

        Parameters:
            value:  Channel grouping to set.
                    0 = 4x2 channels;
                    1 = 2x4 channels;
                    2 = 1x8 channels.
        """
        if value != 2:
            raise ValueError("Unsupported channel grouping")
        self._check_is_open()
        self._set_dev_int("system/awg/channelgrouping", value)

    @rpc_method
    def set_reference_clock_source(self, value: int) -> None:
        """Enable or disable the use of an external reference clock source.

        Parameters:
            value:  Clock source to select.
                    0 = internal reference clock;
                    1 = external 10 MHz reference clock.
        """
        if value not in (0, 1):
            raise ValueError("Unsupported reference clock source")
        self._check_is_open()
        self._set_dev_int("system/clocks/referenceclock/source", value)

    @rpc_method
    def get_reference_clock_status(self) -> int:
        """Return the status of the reference clock.

        Returns:
            0 if the reference clock is locked;
            1 if there was an error locking to the reference clock;
            2 if the device is busy locking to the reference clock.
        """
        self._check_is_open()
        return self._get_dev_int("system/clocks/referenceclock/status")

    @rpc_method
    def set_sample_clock_frequency(self, value: float) -> None:
        """Change the base sample clock frequency.

        Changing the sample clock temporarily interrupts the AWG sequencers.

        Parameters:
            value: New base sample clock frequency in Hz (range 100.0e6 to 2.4e9).
        """
        self._check_is_open()
        self._set_dev_double("system/clocks/sampleclock/freq", value)

    @rpc_method
    def get_sample_clock_status(self) -> int:
        """Return the status of the sample clock.

        Returns:
            0 if the sample clock is valid and locked;
            1 if there was an error adjusting the sample clock;
            2 if the device is busy adjusting the sample clock.
        """
        self._check_is_open()
        return self._get_dev_int("system/clocks/sampleclock/status")

    @rpc_method
    def set_trigger_impedance(self, trigger: int, value: int) -> None:
        """Set the input impedance of a specific trigger input.

        Parameters:
            trigger: Trigger index in the range 0 to 7, corresponding to trigger inputs 1 to 8 on the front panel.
            value:   Impedance setting.
                     0 = 1 kOhm;
                     1 = 50 Ohm.
        """
        if trigger < 0 or trigger >= self.NUM_CHANNELS:
            raise ValueError("Invalid trigger index")
        if value not in (0, 1):
            raise ValueError("Invalid impedance setting")
        self._check_is_open()
        self._set_dev_int("triggers/in/{}/imp50".format(trigger), value)

    @rpc_method
    def set_trigger_level(self, trigger: int, value: float) -> None:
        """Set the trigger voltage level for a specific trigger input.

        Parameters:
            trigger: Trigger index in the range 0 to 7, corresponding to trigger inputs 1 to 8 on the front panel.
            value:   Trigger level in Volt, range -10.0 to +10.0 exclusive.
        """
        if trigger < 0 or trigger >= self.NUM_CHANNELS:
            raise ValueError("Invalid trigger index")
        if not -10.0 < value < 10.0:
            raise ValueError("Invalid trigger level")
        self._check_is_open()
        self._set_dev_double("triggers/in/{}/level".format(trigger), value)

    @rpc_method
    def set_marker_source(self, trigger: int, value: int) -> None:
        """Select the source for a specific marker output.

        Four types of marker sources are supported.
        Values 0 to 3 select internal signals called "AWG Trigger 1" to "AWG Trigger 4".
        These signals are controlled by the sequencer program via explicit calls to "setTrigger()".

        Values 4 to 7 select marker bits from a currently playing waveform.
        To understand the mapping of marker bits to marker channels, consider channels to be grouped
        into pairs. A specific marker output port can only access marker bits from a waveform that
        is playing within the same channel pair.
        Value 4 selects marker bit 1 of the waveform playing on the first channel of the pair.
        Value 5 selects marker bit 2 of the first channel.
        Value 6 selects marker bit 1 of the second channel.
        Value 7 selects marker bit 2 of the second channel.

        Values 8 to 15 select the external trigger signals from trigger inputs 1 to 8.

        Value 17 sets the marker output to a constant high level.
        Value 18 sets the marker output to a constant low level.

        Parameters:
            trigger: Marker index in the range 0 to 7, corresponding to marker outputs 1 to 8 on the front panel.
            value:   Selected marker source as defined above.
        """
        if trigger < 0 or trigger >= self.NUM_CHANNELS:
            raise ValueError("Invalid trigger index")
        if value not in range(16) and value not in (17, 18):
            raise ValueError("Invalid marker source: {}".format(value))
        self._check_is_open()
        self._set_dev_int("triggers/out/{}/source".format(trigger), value)

    @rpc_method
    def set_marker_delay(self, trigger: int, value: float) -> None:
        """Set the output delay for a specific marker output.

        Trigger delay, controls the fine delay of the trigger output. The resolution is 78 ps.

        Parameters:
            trigger: Marker index in the range 0 to 7, corresponding to marker outputs 1 to 8 on the front panel.
            value:   Delay in seconds.
        """
        if trigger < 0 or trigger >= self.NUM_CHANNELS:
            raise ValueError("Invalid trigger index")
        self._check_is_open()
        self._set_dev_double("triggers/out/{}/delay".format(trigger), value)

    @rpc_method
    def set_dig_trigger_source(self, awg: int, trigger: int, value: int) -> None:
        """Select the source of a specific digital trigger channel.

        There are two digital trigger channels which can be acccessed in the
        sequencer program via calls to "waitDigTrigger()" and "playWaveDigTrigger()".

        Parameters:
            awg:     AWG module index (must be 0 in 1x8 channel mode).
            trigger: Digital trigger index in the range 0 to 1, corresponding to digital triggers 1 and 2.
            value:   Trigger source in the range 0 to 7, corresponding to trigger inputs 1 to 8 on the front panel.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if trigger < 0 or trigger > 1:
            raise ValueError("Invalid digital trigger index")
        if value < 0 or value >= self.NUM_CHANNELS:
            raise ValueError("Invalid trigger source")
        self._check_is_open()
        self._set_dev_int("awgs/{}/auxtriggers/{}/channel".format(awg, trigger), value)

    @rpc_method
    def set_dig_trigger_slope(self, awg: int, trigger: int, value: int) -> None:
        """Set the trigger slope of a specific digital trigger channel.

        There are two digital trigger channels which can be acccessed in the
        sequencer program vi calls to "waitDigTrigger()" and "playWaveDigTrigger()".

        Parameters:
            awg:     AWG module index (must be 0 in 1x8 channel mode).
            trigger: Digital trigger index in the range 0 to 1, corresponding to digital triggers 1 and 2.
            value:   Trigger slope.
                     0 = level sensitive (trigger on high signal);
                     1 = trigger on rising edge;
                     2 = trigger on falling edge;
                     3 = trigger on rising and falling edge.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if trigger < 0 or trigger > 1:
            raise ValueError("Invalid digital trigger index")
        if value < 0 or value > 3:
            raise ValueError("Invalid trigger slope")
        self._check_is_open()
        self._set_dev_int("awgs/{}/auxtriggers/{}/slope".format(awg, trigger), value)

    @rpc_method
    def get_output_amplitude(self, awg: int, channel: int) -> float:
        """Get the output scaling factor of the specified channel.

        Parameters:
            awg:     AWG module index in the range 0 to 3.
                     The AWG index selects a pair of output channels.
                     Index 0 selects output channels 1 and 2, index 1 selects channels 2 and 3 etc.
            channel: Channel index in the range 0 to 1, selecting the first or second output channel
                     within the selected channel pair.
        Returns:
            The output scaling factor, which is a dimensionless scaling factor applied to the digital signal.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if channel < 0 or channel >= self.NUM_CHANNELS_PER_AWG:
            raise ValueError("Invalid channel index")
        self._check_is_open()
        return self._get_dev_double(f"awgs/{awg}/outputs/{channel}/amplitude")

    @rpc_method
    def set_output_amplitude(self, awg: int, channel: int, value: float) -> None:
        """Set the output scaling factor of the specified channel.

        The amplitude is a dimensionless scaling factor applied to the digital signal.

        Parameters:
            awg:     AWG module index in the range 0 to 3.
                     The AWG index selects a pair of output channels.
                     Index 0 selects output channels 1 and 2, index 1 selects channels 2 and 3 etc.
            channel: Channel index in the range 0 to 1, selecting the first or second output channel
                     within the selected channel pair.
            value:   Amplitude scale factor.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if channel < 0 or channel >= self.NUM_CHANNELS_PER_AWG:
            raise ValueError("Invalid channel index")
        self._check_is_open()
        self._set_dev_double("awgs/{}/outputs/{}/amplitude".format(awg, channel), value)

    @rpc_method
    def get_output_channel_hold(self, awg: int, channel: int) -> int:
        """Get whether the last sample is held for the specified channel.

        Parameters:
            awg:     AWG module index in the range 0 to 3.
                     The AWG index selects a pair of output channels.
                     Index 0 selects output channels 1 and 2, index 1 selects channels 2 and 3 etc.
            channel: Channel index in the range 0 to 1, selecting the first or second output channel
                     within the selected channel pair.
        Returns:
            0 if not last sample is not held;
            1 if last sample is held.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if channel < 0 or channel >= self.NUM_CHANNELS_PER_AWG:
            raise ValueError("Invalid channel index")
        self._check_is_open()
        return self._get_dev_int(f"awgs/{awg}/outputs/{channel}/hold")

    @rpc_method
    def set_output_channel_hold(self, awg: int, channel: int, value: int) -> None:
        """Set whether the last sample should be held for the specified channel.

        Parameters:
            awg:     AWG module index in the range 0 to 3.
                     The AWG index selects a pair of output channels.
                     Index 0 selects output channels 1 and 2, index 1 selects channels 2 and 3 etc.
            channel: Channel index in the range 0 to 1, selecting the first or second output channel
                     within the selected channel pair.
            value:   Hold state; 0 = False, 1 = True.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if channel < 0 or channel >= self.NUM_CHANNELS_PER_AWG:
            raise ValueError("Invalid channel index")
        if value not in (0, 1):
            raise ValueError("Invalid hold state")
        self._check_is_open()
        self._set_dev_int(f"awgs/{awg}/outputs/{channel}/hold", value)

    @rpc_method
    def get_output_channel_on(self, output_channel: int) -> int:
        """Get the specified wave output channel state: 0 = output off, 1 = output on.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        self._check_is_open()
        return self._get_dev_int("sigouts/{}/on".format(output_channel))

    @rpc_method
    def set_output_channel_on(self, output_channel: int, value: int) -> None:
        """Set the specified wave output channel on or off.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
            value:          Output state; 0 = output off, 1 = output on.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        if value not in (0, 1):
            raise ValueError("Invalid on/off state")
        self._check_is_open()
        self._set_dev_int("sigouts/{}/on".format(output_channel), value)

    @rpc_method
    def set_output_channel_range(self, output_channel: int, value: float) -> None:
        """Set output voltage range of the specified wave output channel.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
            value:          Output range in Volt. The instrument selects the next higher available range.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        if value < 0 or value > 5.0:
            raise ValueError("Invalid output range")
        self._check_is_open()
        self._set_dev_double("sigouts/{}/range".format(output_channel), value)

    @rpc_method
    def set_output_channel_offset(self, output_channel: int, value: float) -> None:
        """Set the DC offset voltage for the specified wave output channel.

        The DC offset is only active in amplified mode, not in direct mode.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
            value:          Offset in Volt, in range -1.25 to +1.25 V.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        if not -1.25 <= value <= 1.25:
            raise ValueError("Invalid offset value")
        self._check_is_open()
        self._set_dev_double("sigouts/{}/offset".format(output_channel), value)

    @rpc_method
    def get_output_channel_delay(self, output_channel: int) -> float:
        """Return output delay for fine alignment of the the specified wave output channel.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.

        Returns:
            Delay in seconds.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        self._check_is_open()
        return self._get_dev_double("sigouts/{}/delay".format(output_channel))

    @rpc_method
    def set_output_channel_delay(self, output_channel: int, value: float) -> None:
        """Set output delay for fine alignment of the the specified wave output channel.

        Changing the delay setting may take several seconds.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
            value:          Delay in seconds, range 0 to 26e-9.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        if not 0 <= value < 26e-9:
            raise ValueError("Invalid delay setting")
        self._check_is_open()
        self._set_dev_double("sigouts/{}/delay".format(output_channel), value)

    @rpc_method
    def set_output_channel_direct(self, output_channel: int, value: int) -> None:
        """Enable or disable the direct output path for the specified wave output channel.

        The direct output path bypasses the output amplifier and offset circuits,
        and fixes the output range to 800 mV.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
            value:          1 to enable the direct output path, 0 to disable.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        if value not in (0, 1):
            raise ValueError("Invalid value")
        self._check_is_open()
        self._set_dev_int("sigouts/{}/direct".format(output_channel), value)

    @rpc_method
    def set_output_channel_filter(self, output_channel: int, value: int) -> None:
        """Enable or disable the analog output filter for the specified wave output channel.

        Parameters:
            output_channel: Channel index in the range 0 to 7, corresponding to wave outputs 1 to 8 on the front panel.
            value:          1 to enable the output filter, 0 to disable.
        """
        if output_channel < 0 or output_channel >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")
        if value not in (0, 1):
            raise ValueError("Invalid value")
        self._check_is_open()
        self._set_dev_int("sigouts/{}/filter".format(output_channel), value)

    @rpc_method
    def set_dio_mode(self, mode: int) -> None:
        """Set the DIO control mode.

        Available mode are:
            0: manual:                  control of DIO bits via LabOne user interface.
            1: awg_sequencer_commands:  control of DIO bits from sequencer code and forward set values to sequencer
                                        (interface synced to 150 MHz).
            2: dio_codeword             same as (1), but interface is synced to 50 MHz.
            3: qccs                     map DIO onto ZSync interface and make it available to the sequencer.

        Parameters:
            mode:   mode to use (0, 1, 2 or 3).
        """
        if not 0 <= mode <= 3:
            raise ValueError("Invalid DIO mode")
        self._check_is_open()
        self._set_dev_int("dios/0/mode", mode)

    @rpc_method
    def set_dio_drive(self, mask: int) -> None:
        """Define input/output direction of DIO signals.

        Parameters:
            mask:   4-bit mask where each bit configures the direction of
                    a group of 8 DIO signals. Bit value 0 sets the DIO signals
                    to input mode, value 1 sets the signals to output mode.
                    Bit 0 corresponds to the least significant byte of the DIO bus, etc.
        """
        if not 0 <= mask <= 15:
            raise ValueError("Invalid value for mask")
        self._check_is_open()
        self._set_dev_int("dios/0/drive", mask)

    @rpc_method
    def set_dio_strobe_index(self, awg: int, value: int) -> None:
        """Select the DIO index to be used as a trigger for playback.

        The sequencer program uses this trigger by calling "playWaveDIO()".

        Parameters:
            awg:    AWG module index (must be 0 in 1x8 channel mode).
            value:  DIO bit index in the range 0 to 31.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if value < 0 or value > 31:
            raise ValueError("Invalid DIO bit index")
        self._check_is_open()
        self._set_dev_int("awgs/{}/dio/strobe/index".format(awg), value)

    @rpc_method
    def set_dio_strobe_slope(self, awg: int, value: int) -> None:
        """Select the signal edge that should activate the strobe trigger.

        Parameters:
            awg:    AWG module index (must be 0 in 1x8 channel mode).
            value:  Slope type.
                    0 = off;
                    1 = trigger on rising edge;
                    2 = trigger on falling edge;
                    3 = trigger on rising and falling edge.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if value < 0 or value > 3:
            raise ValueError("Invalid slope")
        self._check_is_open()
        self._set_dev_int("awgs/{}/dio/strobe/slope".format(awg), value)

    @rpc_method
    def get_user_register(self, awg: int, reg: int) -> int:
        """Return the value of the specified user register.

        Parameters:
            awg:    AWG module index (must be 0 in 1x8 channel mode).
            reg:    Register index in the range 0 to 15.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if not 0 <= reg <= 15:
            raise ValueError("Invalid register index")
        self._check_is_open()
        return self._get_dev_int("awgs/{}/userregs/{}".format(awg, reg))

    @rpc_method
    def set_user_register(self, awg: int, reg: int, value: int) -> None:
        """Change the value of the specified user register.

        Parameters:
            awg:    AWG module index (must be 0 in 1x8 channel mode).
            reg:    Register index in the range 0 to 15.
            value:  Integer value to write to the register.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        if not 0 <= reg <= 15:
            raise ValueError("Invalid register index")
        self._check_is_open()
        self._set_dev_int("awgs/{}/userregs/{}".format(awg, reg), value)

    @rpc_method
    def get_awg_module_enabled(self, awg: int) -> int:
        """Return the current enable status of the AWG module.

        Parameters:
            awg:    AWG module index (must be 0 in 1x8 channel mode).

        Returns:
            1 if the AWG sequencer is currently running;
            0 if the AWG sequencer is not running.
        """
        if awg < 0 or awg >= self.NUM_AWGS:
            raise ValueError("Invalid AWG index")
        self._check_is_open()
        return self._get_dev_int("awgs/{}/enable".format(awg))

    @rpc_method
    def set_awg_module_enabled(self, value: int) -> None:
        """Enable or disable the AWG module.

        Enabling the AWG starts execution of the currently loaded sequencer program.
        Uploading a sequencer program or waveform is only possible while the AWG module is disabled.

        Parameters:
            value: 1 to enable the AWG module, 0 to disable.
        """
        if value not in (0, 1):
            raise ValueError("Invalid value")
        self._check_is_open()
        assert self._awg_module is not None
        self._awg_module.set("awgModule/awg/enable", value)

    # Obsolete method names for backward compatibility.
    setChannelGrouping = set_channel_grouping
    setReferenceClockSource = set_reference_clock_source
    setTriggerImpedance = set_trigger_impedance
    setTriggerSource = set_marker_source
    setAuxTriggerSlope = set_dig_trigger_slope
    setOutputAmplitude = set_output_amplitude
    setOutputChannelOn = set_output_channel_on
    setOutputChannelRange = set_output_channel_range
    setDioDrive = set_dio_drive
    setDioStrobeIndex = set_dio_strobe_index
    setDioStrobeSlope = set_dio_strobe_slope
    setUserRegister = set_user_register
    setAwgModuleEnabled = set_awg_module_enabled



#! /usr/bin/env python3
import logging
import time
import unittest

import qmi
from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.context import QMI_Context, QMI_Instrument
from qmi.core.exceptions import QMI_RuntimeException, QMI_Exception, QMI_TaskRunException, QMI_UsageException, \
    QMI_InvalidOperationException, QMI_ConfigurationException, QMI_UnknownNameException
from qmi.core.rpc import rpc_method
from qmi.core.task import QMI_Task
from qmi.core.pubsub import QMI_Signal, QMI_SignalReceiver


SPEED_OF_LIGHT = 2998792458


class MyContextTestClass(QMI_Instrument):

    def __init__(self, context: QMI_Context, name: str) -> None:
        super().__init__(context, name)
        self.frequency = 2.0

    @rpc_method
    def get_wavelength(self) -> float:
        """This is some get_method."""
        return SPEED_OF_LIGHT / self.frequency

    @rpc_method
    def set_frequency(self, f: float) -> None:
        """This is some set_method."""
        if f <= 0.0:
            raise ValueError("Value must be positive!")

        self.frequency = f


class TestMultipleContextsSimple(unittest.TestCase):

    def test_multiple_contexts(self):
        """ Test that multiple contexts can be started simultaneously."""
        c1 = QMI_Context("c1", config=None)
        c2 = QMI_Context("c2", config=None)
        c3 = QMI_Context("c3", config=None)

        c1.start()
        c2.start()
        c3.start()

        time.sleep(1.0)
        # Get info from the contexts and see that the name is correct and that they are all active
        contexts = [c1, c2, c3]
        for e, context in enumerate(contexts):
            name = context.info().split("\'")[1]
            self.assertEqual(name, "c{}".format(e + 1))
            self.assertTrue(context._active)

        for con in c1.get_rpc_object_descriptors():
            self.assertEqual(con.address.context_id, "c1")
            self.assertEqual(con.category, "context")

        # Stop the contexts and see that they all stop
        c1.stop()
        c2.stop()
        c3.stop()
        for context in contexts:
            self.assertFalse(context._active)

    def test_invalid_context_names(self):
        """Test that invalid object names raise exceptions."""
        too_long = ""
        for _ in range(64):
            too_long += "a"

        with self.assertRaises(QMI_UsageException):
            QMI_Context("#@%^", config=None)

        with self.assertRaises(QMI_UsageException):
            QMI_Context(too_long, config=None)


class TestMultipleContextsSpecial(unittest.TestCase):

    def setUp(self):
        config = CfgQmi(
            contexts={
                "c1": CfgContext(host="localhost", tcp_server_port=12345),
                "c2": CfgContext(host="localhost", tcp_server_port=12346),
                "c3": CfgContext(tcp_server_port=0)
            }
        )
        self.c1 = QMI_Context("c1", config=config)
        self.c2 = QMI_Context("c2", config=config)
        self.c3 = QMI_Context("c3", config=None)

    def tearDown(self) -> None:
        # Stop the contexts
        if self.c1._active:
            self.c1.stop()

        if self.c2._active:
            self.c2.stop()

        if self.c3._active:
            self.c3.stop()

    def test_multiple_context_configs(self):
        """Try to connect to this peer without giving a peer address. The config has host and port so it should work.
        Then try to connect to this peer without giving a peer address. The config is None, so exception should be
        raised."""
        self.c1.start()
        self.c2.start()
        self.c3.start()

        self.c1.connect_to_peer("c2")  # This should work fine as config has host name and port.
        with self.assertRaises(QMI_UnknownNameException):
            self.c3.connect_to_peer("c2")  # This should fail as for c3 config is None.

    def test_get_tcp_server_port(self):
        """See that we can get the correct port number form active context and 0 from non-active."""
        self.c1.start()
        port1 = self.c1.get_tcp_server_port()
        self.assertEqual(port1, 12345)

        port2 = self.c2.get_tcp_server_port()
        self.assertEqual(port2, 0)


class TestMultipleContextsPeers(unittest.TestCase):

    def test_connect_to_peers(self):
        """Try to connect to multiple peers with the `start()` command."""
        config = {
            "c1": {"host": "localhost", "tcp_server_port": 12345},
            "c2": {"host": "localhost", "tcp_server_port": 12346},
            "c3": {"tcp_server_port": 0, "connect_to_peers": ["c1", "c2"]}
        }

        c1 = QMI_Context("c1", config=CfgQmi(contexts={"c1": CfgContext(host="localhost", tcp_server_port=12345)}))
        c2 = QMI_Context("c2", config=CfgQmi(contexts={"c2": CfgContext(host="localhost", tcp_server_port=12346)}))
        c1.start()
        c2.start()

        qmi.start("c3", context_cfg=config)
        contexts = qmi.get_configured_contexts()
        expect_contexts = {
            "c1": CfgContext(host="localhost", tcp_server_port=12345),
            "c2": CfgContext(host="localhost", tcp_server_port=12346),
            "c3": CfgContext(tcp_server_port=0, connect_to_peers=["c1", "c2"])
        }
        qmi.stop()
        c1.stop()
        c2.stop()

        self.assertEqual(contexts, expect_contexts)


class TestQMIContextClass(unittest.TestCase):

    def setUp(self):
        # Start two contexts.
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0),
                "c2": CfgContext(tcp_server_port=0)
            }
        )

        c1 = QMI_Context("c1", config)
        c1.start()
        c1_port = c1.get_tcp_server_port()

        c2 = QMI_Context("c2", config)
        c2.start()
        c2_port = c2.get_tcp_server_port()

        # Note: Connecting to "localhost" on Windows can give 1 second delay
        # because IPv6 is attempted before IPv4.
        # This problem does not occur when connecting to "127.0.0.1".

        # Connect c1 to c2.
        c1.connect_to_peer("c2", "localhost:{}".format(c2_port))

        # Connect c2 to c1.
        c2.connect_to_peer("c1", "127.0.0.1:{}".format(c1_port))

        self.c1 = c1
        self.c2 = c2

    def tearDown(self):

        if self.c1._active:
            self.c1.stop()

        self.c2.stop()

        self.c1 = None
        self.c2 = None

    def test_has_peer_context(self):
        """See that a context with specific name is connected as a peer."""
        has_it = self.c1.has_peer_context("c2")
        self.assertTrue(has_it)

    def test_connect_context_name_mismatch(self):
        """See that an exception is raised if the context names do not exist or do not match"""
        c2_port = self.c2.get_tcp_server_port()
        with self.assertRaises(QMI_RuntimeException):
            self.c1.connect_to_peer("aap", "127.0.0.1:{}".format(c2_port))

    def test_connect_to_peer_without_address_and_no_config(self):
        """Try to connect to this peer without giving a peer address and no host. See that exception is raised."""
        with self.assertRaises(QMI_ConfigurationException):
            self.c1.connect_to_peer("c2")

    def test_connect_duplicate_peer_ignore(self):
        """Do nothing as we are already connected to this peer and we choose to ignore it."""
        c2_port = self.c2.get_tcp_server_port()
        self.c1.connect_to_peer("c2", "localhost:{}".format(c2_port), ignore_duplicate=True)

    def test_connect_inactive_context_cannot_connect_to_peer(self):
        """Raise an exception as we are trying to connect an inactive context"""
        self.c1.stop()
        c2_port = self.c2.get_tcp_server_port()
        with self.assertRaises(QMI_InvalidOperationException):
            self.c1.connect_to_peer("c2", "localhost:{}".format(c2_port))

    def test_disconnect_inactive_context_raises_exception(self):
        """Raise an exception as we are trying to disconnect an inactive context"""
        self.c1.stop()
        with self.assertRaises(QMI_InvalidOperationException):
            self.c1.disconnect_from_peer("c2")

    def test_connect_duplicate_peer_raises_exception(self):
        """Raise an exception as we are already connected to this peer"""
        c2_port = self.c2.get_tcp_server_port()
        with self.assertRaises(QMI_UsageException):
            self.c1.connect_to_peer("c2", "localhost:{}".format(c2_port))


class TestMultipleUsersToInstrument(unittest.TestCase):

    def setUp(self):
        # Suppress logging.
        logging.getLogger("qmi.core.rpc").setLevel(logging.CRITICAL)

        # Start two contexts with the first one creating the "instrument".
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0),
                "c2": CfgContext(tcp_server_port=0)
            }
        )
        self.c1 = QMI_Context("c1", config=config)
        self.c1.start()
        self.c2 = QMI_Context("c2", config=config)
        self.c2.start()

        self.instr_proxy_server = self.c1.make_instrument("instr", MyContextTestClass)

        # Connect second context to first.
        c1_port = self.c1.get_tcp_server_port()
        self.c2.connect_to_peer("c1", f"localhost:{c1_port}")
        self.instr_proxy_client = self.c2.get_instrument("c1.instr")

    def tearDown(self):
        self.c1.stop()
        self.c2.stop()

        self.c1 = None
        self.c2 = None

        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_01_get_instrument(self):
        """Test that second proxy to instrument can send commands without problems."""
        frequency = 2.0
        self.instr_proxy_client.set_frequency(frequency)

        wl = self.instr_proxy_server.get_wavelength()
        self.assertEqual(wl, SPEED_OF_LIGHT / frequency)

    def test_02_check_acquire_and_unlock_object(self):
        """Test that the acquire, is locked enquiry, and release work."""
        # Test initial state is False
        ans = self.instr_proxy_client.is_locked()
        self.assertFalse(ans)
        # Test that acquiring lock changes state to True
        success = self.instr_proxy_client.lock()
        self.assertTrue(success)
        ans = self.instr_proxy_client.is_locked()
        self.assertTrue(ans)
        # Test that releasing the lock changes state back to False
        success = self.instr_proxy_client.unlock()
        self.assertTrue(success)
        ans = self.instr_proxy_client.is_locked()
        self.assertFalse(ans)

    def test_03_lock_object_on_first_proxy(self):
        """Test that second proxy to instrument fails to get data after first proxy acquires a lock."""
        # Set lock
        success = self.instr_proxy_client.lock()
        self.assertTrue(success)
        # See that the second proxy fails to command the instrument
        with self.assertRaises(QMI_RuntimeException):
            self.instr_proxy_server.set_frequency(1.0)

    def test_04_acquire_and_unlock_object_on_first_proxy(self):
        """Test that second proxy to instrument can send commands data after first proxy releases the lock."""
        # set lock and check
        self.instr_proxy_client.lock()
        ans = self.instr_proxy_server.is_locked()
        self.assertTrue(ans)
        # release the lock and check
        self.instr_proxy_client.unlock()
        ans = self.instr_proxy_server.is_locked()
        self.assertFalse(ans)
        # See that it works again
        frequency = 4.0
        self.instr_proxy_server.set_frequency(frequency)
        wl = self.instr_proxy_client.get_wavelength()
        self.assertEqual(wl, SPEED_OF_LIGHT / frequency)


class TestMultipleUsersToInstrumentWithLock(unittest.TestCase):

    def setUp(self):
        # Suppress logging.
        logging.getLogger("qmi.core.rpc").setLevel(logging.CRITICAL)

        # Start two contexts with the first one creating the "instrument".
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0),
                "c2": CfgContext(tcp_server_port=0)
            }
        )
        self.c1 = QMI_Context("c1", config=config)
        self.c1.start()
        self.c2 = QMI_Context("c2", config=config)
        self.c2.start()

        self.instr_proxy_server = self.c1.make_rpc_object("instr", MyContextTestClass)

        # Connect second context to first.
        c1_port = self.c1.get_tcp_server_port()
        self.c2.connect_to_peer("c1", f"localhost:{c1_port}")
        self.instr_proxy_client = self.c2.get_rpc_object_by_name("c1.instr")

    def tearDown(self):
        self.c1.stop()
        self.c2.stop()

        self.c1 = None
        self.c2 = None

        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_01_check_command_before_lock_works(self):
        """Test that first proxy to instrument can send commands without problems."""
        frequency = 2.0
        self.instr_proxy_client.set_frequency(frequency)
        wl = self.instr_proxy_client.get_wavelength()
        self.assertEqual(wl, SPEED_OF_LIGHT / frequency)

    def test_02_cannot_lock_object_for_already_locked_instrument(self):
        """Test that first proxy to instrument fails to acquire lock after second proxy acquires a lock."""
        self.instr_proxy_server.lock()
        success = self.instr_proxy_client.lock()
        self.assertFalse(success)

    def test_03_cannot_unlock_object_for_already_locked_instrument(self):
        """Test that first proxy to instrument fails to release lock after second proxy acquires a lock."""
        self.instr_proxy_server.lock()
        success = self.instr_proxy_client.unlock()
        self.assertFalse(success)

    def test_04_command_after_second_proxy_lock_raises_exception(self):
        """Test that first proxy to instrument fails to set/get data after second proxy acquires a lock."""
        self.instr_proxy_server.lock()

        with self.assertRaises(QMI_RuntimeException):
            self.instr_proxy_client.set_frequency(1.0)

        with self.assertRaises(QMI_RuntimeException):
            self.instr_proxy_client.get_wavelength()

    def test_05_force_release_locked_instrument(self):
        """Test that first proxy's lock can be released forcefully and is available for the second afterwards."""
        success = self.instr_proxy_client.lock()
        self.assertTrue(success)
        self.assertTrue(self.instr_proxy_server.is_locked())

        # Force unlock.
        self.instr_proxy_server.force_unlock()
        self.assertFalse(self.instr_proxy_server.is_locked())

        # And check that trying to lock it no longer fails.
        success = self.instr_proxy_server.lock()
        self.assertTrue(success)
        self.assertTrue(self.instr_proxy_server.is_locked())


class MyContextTaskController(QMI_Task):
    sig_sample = QMI_Signal([float])
    lock = False

    def __init__(self, task_runner, name, instrument, sample_time):
        super().__init__(task_runner, name)

        self._instr = instrument
        self._sample_time = sample_time

    def run(self):
        """Main loop."""
        try:
            self._instr.lock()

            # Sweep through optical range of wavelengths (350-700nm)
            blue_frequency = SPEED_OF_LIGHT / 350e-9
            red_frequency = SPEED_OF_LIGHT / 700e-9
            sweep_steps = 8
            step_size = (red_frequency - blue_frequency) / (sweep_steps - 1)
            for step in range(sweep_steps):
                frequency = blue_frequency + step * step_size
                self._instr.set_frequency(frequency)
                wl = self._instr.get_wavelength()
                self.sig_sample.publish(wl)
                time.sleep(self._sample_time)

        except QMI_Exception as exc:
            raise QMI_Exception from exc

        finally:
            if self._instr.is_locked():
                self._instr.unlock()


class TestTasksAndUserToInstrumentWithLock(unittest.TestCase):

    def setUp(self):
        # Suppress logging.
        logging.getLogger("qmi.core.rpc").setLevel(logging.CRITICAL)
        logging.getLogger("qmi.core.task").setLevel(logging.ERROR)

        # Start a context with creating an "instrument".
        self.c1 = QMI_Context("instr_server", config=None)
        self.c1.start()
        self.instr = self.c1.make_instrument("instr", MyContextTestClass)

        # We need a client to talk to the created instrument (using self.instr directly does not work)
        client = self.c1.get_instrument("instr_server.instr")

        # Then prepare a task within that context.
        self.controller = self.c1.make_task("controller", MyContextTaskController, client, 0.1)

        # Create a signal subscriber
        self.receiver = QMI_SignalReceiver()
        self.publisher_proxy = self.c1.get_task("instr_server.controller")
        self.publisher_proxy.sig_sample.subscribe(self.receiver)
        self.wait_timeout = 0.1

    def tearDown(self) -> None:
        # Close task and QMI
        try:
            if self.controller.is_running():
                self.controller.stop()
                self.controller.join()

        except QMI_TaskRunException:
            pass

        finally:
            self.c1.stop()
            logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)
            logging.getLogger("qmi.core.task").setLevel(logging.NOTSET)

    def test_01_task_runs_normally(self):
        """Test that the task runs as expected."""
        expected = [3.5e-07, 3.77e-07, 4.08e-07, 4.45e-07, 4.9e-07, 5.44e-07, 6.13e-07, 7e-07]
        self.controller.start()
        for s in range(len(expected)):
            # Wait that data is produced and check that we are still running.
            while not self.receiver.has_signal_ready() and self.controller.is_running():
                time.sleep(self.wait_timeout)

            signal = self.receiver.get_next_signal(timeout=self.wait_timeout)
            self.assertEqual(expected[s], round(signal.args[0], 9))

    def test_02_task_cannot_run_due_to_instrument_locked(self):
        """The instrument is locked before running the task, forcing the task to except."""
        self.instr.lock()

        with self.assertRaises(QMI_TaskRunException):
            self.controller.start()
            time.sleep(self.wait_timeout)
            self.controller.stop()
            self.controller.join()

        # See that controller did not do anything to set frequencies
        wl = self.instr.get_wavelength()
        self.assertEqual(SPEED_OF_LIGHT / 2.0, wl)

    def test_03_task_runs_with_locking_the_instrument(self):
        """The task locks the instrument and runs normally. The instrument cannot be communicated
        from the regular context."""
        self.controller.start()
        time.sleep(self.wait_timeout)
        with self.assertRaises(QMI_RuntimeException):
            self.instr.set_frequency(2.0)


if __name__ == "__main__":
    unittest.main()



"""Remote Procedure Call mechanism for QMI.

This module provides support for Remote Procedure Calls: calling methods
on an object that exists in a different QMI context.

Defining RPC methods
####################

Only classes that inherit `QMI_RpcObject` can define RPC methods.
Within such a class, use the `@rpc_method` decorator to mark a method as
RPC-callable. Note that `QMI_Instrument` inherits `QMI_RpcObject`, therefore
instrument drivers are able to define RPC methods.

For example, the following fragment defines an RPC-callable method `MyClass.square`::

  class MyClass(QMI_RpcObject):

      def __init__(self, context, name, ...):
          super().__init__(context, name)

      @rpc_method
      def square(self, x):
          return x * x

Some restrictions apply to the types of data that can be passed in and out
of RPC methods: Parameters passed to an RPC method and return values or
exceptions produced by an RPC method must be `pickle-able`, i.e. the
Python `pickle` module must be able to serialize the data types used.

* Built-in types are generally pickle-able.
* Custom defined value-like types such as enums, named tuples and exceptions,
  are generally pickle-able provided that the type definition can be imported
  by the receiving program. This implies that such types must be defined
  in a Python module, not in a top-level Python script.
* Numpy arrays are pickle-able.
* Other kinds of custom defined classes are generally not pickle-able.

Calling RPC methods
###################

Classes that inherit `QMI_RpcObject` are instantiated via the QMI context.
The context then returns a `proxy` for the new object instance.

RPC methods can be called by simply invoking a method with the same name
on the proxy. The RPC mechanism translates this into a call to the real method
on the real object. The return value from the real method is passed back as
the return value from the proxy call.

For example::

    proxy = qmi.context().make_rpc_object("my_object", MyClass, ...)
    y = proxy.square(5)

It is also possible to create a proxy for an object in a different QMI context.
In this case, calling a method on the proxy will cause the method with the same
name to run in the remote context::

    proxy = qmi.context().get_rpc_object_by_name("other_context.my_object")
    y = proxy.square(5)

Locking RPC objects
###################

Class that inherit from `QMI_RpcObject` are lockable. This means that only the
proxy that owns the lock (i.e. that acquired it) can invoke RPC methods. Other
proxies will not be able to call any RPC method until the object is unlocked.
An example use case is where you have a script running to do some measurement
and you want to avoid that a GUI or automatic calibration interferes with that
measurement: in that case you can have the script lock the instruments.

Example::

    proxy = qmi.context().get_rpc_object_by_name("other_context.my_object")
    proxy.lock()  # object is now locked
    # do things ...
    proxy.unlock()

You can query if an object is locked via `is_locked()`. If for whatever reason
an object is locked and the proxy that owns the locked no longer exists, you
can force unlock the object via `force_unlock()`; use this with care!

The normal proxy locking is available only within the same context for the same proxy.
Since QMI V0.29.1 it is also possible to lock and unlock with a custom token from
other proxies as well, as long as the contexts for the other proxies have the same name.

Example 1::
        # Two proxies in same context.
        proxy1 = context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = context.get_rpc_object_by_name("my_context.my_object")
        custom_token = "thisismineallmine"

        proxy1.lock(lock_token=custom_token)
        proxy2.is_locked()  # Returns True
        proxy2.unlock(lock_token=custom_token)
        proxy2.is_locked()  # Returns False

Example 2::
        # Three proxies in different contexts. The first one serves as an "object provider".
        c1 = QMI_Context("c1", config)
        c1.start()
        c1_port = c1.get_tcp_server_port()
        c1_address = "localhost:{}".format(c1_port)
        # Instantiate class in context c1 as object provider.
        proxy1 = c1.make_rpc_object("tc1", MyRpcTestClass)

        # Now make another context, and get a proxy to the object
        custom_token = "block"
        c2 = QMI_Context("c2", config)
        c2.start()
        c2.connect_to_peer("c1", c1_address)
        proxy2 = c2.get_rpc_object_by_name("c1.tc1")
        proxy2.is_locked()  # Returns False
        proxy2.lock(lock_token=custom_token)
        proxy2.is_locked()  # Returns True
        # And then close it
        c2.stop()

        # Then start third context with the same name as second context, obtain the object and unlock it.
        c3 = QMI_Context("c2", config)
        c3.start()
        c3.connect_to_peer("c1", c1_address)
        proxy3 = c3.get_rpc_object_by_name("c1.tc1")
        proxy3.is_locked()  # Returns True
        proxy3.unlock()  # Returns False as it fails without the correct token
        proxy3.unlock(lock_token=custom_token)  # Now succeeds and returns True
        proxy3.is_locked()  # Returns False
        # And then close it
        c3.stop()

        # Then finally stop the "object provider"
        c1.stop()

Reference
#########
"""

import inspect
import logging
import threading
import time
import enum
from abc import ABCMeta
from collections import deque

import typing
from typing import Any, Callable, List, NamedTuple, Optional, Type, TypeVar, Union

from qmi.core.exceptions import (
    QMI_RuntimeException,
    QMI_UsageException,
    QMI_MessageDeliveryException,
    QMI_RpcTimeoutException,
    QMI_UnknownRpcException)
from qmi.core.messaging import (
    QMI_Message, QMI_RequestMessage, QMI_ReplyMessage, QMI_ErrorReplyMessage,
    QMI_MessageHandler, QMI_MessageHandlerAddress)
from qmi.core.pubsub import SignalDescription, QMI_Signal, QMI_RegisteredSignal, QMI_SignalSubscriber
from qmi.core.thread import QMI_Thread
from qmi.core.util import is_valid_object_name


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Unbound type variable for type annotations.
_T = TypeVar("_T")


class RpcConstantDescriptor(NamedTuple):
    """Description of an RPC constant.

    Attributes:
        name: Name of the constant.
        value: Value of the constant.
    """
    name: str
    value: Any


class RpcMethodDescriptor(NamedTuple):
    """Description of an RPC method.

    Attributes:
        name: Name of the RPC method.
        signature: String representation of the signature of the RPC method,
            including type annotations.
        docstring: Docstring of the RPC method.
    """
    name: str
    signature: str
    docstring: str


class RpcSignalDescriptor(NamedTuple):
    """Description of a QMI signal.

    Attributes:
        name: Name of the signal.
        arg_types: String representation of the list of argument types.
    """
    name: str
    arg_types: str


class RpcInterfaceDescriptor(NamedTuple):
    """Description of the subset of the interface of an RPC object class that
    can be accessed via RPC. This includes the methods marked using the
    `@rpc_method` decorator and the signals declared by the RPC object class or
    a delegate class.

    Attributes:
        rpc_class_module: Name of the module in which the RPC object class was
            defined.
        rpc_class_name: Name of the RPC object class.
        rpc_class_docstring: Docstring of the RPC object class.
        constants: A list of constant descriptors for the RPC constants declared
            by the RPC object class.
        methods: A list of method descriptors for the RPC methods declared by
            the RPC object class.
        signals: A list of signal descriptors for the signals declared by the
            RPC object class or a delegate class.
    """
    rpc_class_module: str
    rpc_class_name: str
    rpc_class_docstring: Optional[str]
    constants: List[RpcConstantDescriptor]
    methods: List[RpcMethodDescriptor]
    signals: List[RpcSignalDescriptor]


class RpcObjectDescriptor(NamedTuple):
    """Description of an RPC object instance.

    Attributes:
        address: Unique address of the RPC object.
        category: Free-form name of the category of objects this RPC object
            belongs to.
        interface: Description of the subset of the interface of the RPC object
            that can be accessed via RPC, including signals.
    """
    address: QMI_MessageHandlerAddress
    category: Optional[str]
    interface: RpcInterfaceDescriptor


class QMI_LockTokenDescriptor(NamedTuple):
    """Unique lock token that is used to lock/unlock RPC objects.

    Attributes:
        context_id: Name of the context that owns the RPC proxy that requested the lock.
        token:      Unique token.
    """
    context_id: str
    token: str


class QMI_RpcFutureState(enum.Enum):
    """Possible states of an QMI_RpcFuture instance."""
    NO_RESULT_YET = 1
    RESULT_IS_VALUE = 2
    RESULT_IS_EXCEPTION = 3
    OBJECT_IS_LOCKED = 4


class QMI_LockRpcAction(enum.Enum):
    """Actions that can be performed on a lock."""
    ACQUIRE = 1
    RELEASE = 2
    FORCE_RELEASE = 3
    QUERY = 4


# These "token" is returned when a lock is queried by a proxy that does not own the lock.
ACCESS_DENIED_TOKEN_PLACEHOLDER = "__ACCESS_DENIED__"
OBJECT_LOCKED_TOKEN_PLACEHOLDER = "__OBJECT_LOCKED__"


class QMI_LockRpcRequestMessage(QMI_RequestMessage):
    """Message sent by an RPC client to interact with the lock state of a remote object.

    See `QMI_LockRpcReplyMessage` for how to interpret the reply to a request.

    Attributes:
        lock_token:     The unique token to use for the lock.
        lock_action:    The action to be performed on the lock state.
    """
    __slots__ = ("lock_token", "lock_action")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 lock_token: Optional[QMI_LockTokenDescriptor],
                 lock_action: QMI_LockRpcAction
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.lock_token = lock_token
        self.lock_action = lock_action


class QMI_LockRpcReplyMessage(QMI_ReplyMessage):
    """Message sent back to an RPC client with the result of the action.

    The reply only contains a lock token (an actual token or a placeholder), indicating if the request was successful
    and what the state of the object lock is after the request.

    Specifically:
     - if you requested a lock and the returned token matches your token, you now own the lock;
     - if your requested to unlock and the returned token is None, the unlock was successful;
     - if your queried the lock status and the returned token is not None when the object is locked and None if the
       object is unlocked;
     - in all other cases, the request was denied.

    Attributes:
        lock_token: The unique token that locked the object.
    """
    __slots__ = ("lock_token",)

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 request_id: str,
                 lock_token: Optional[QMI_LockTokenDescriptor]
                 ) -> None:
        super().__init__(source_address, destination_address, request_id)
        self.lock_token = lock_token


class QMI_MethodRpcRequestMessage(QMI_RequestMessage):
    """Message sent by an RPC client to invoke a remote method.

    Attributes:
        method_name:    Name of the method to invoke.
        method_args:    Tuple of positional arguments to the method.
        method_kwargs:  Dictionary of keyword arguments to the method.
        lock_token:     The unique token to use for the lock.
    """
    __slots__ = ("method_name", "method_args", "method_kwargs", "lock_token")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 method_name: str,
                 method_args: tuple,
                 method_kwargs: dict,
                 lock_token: Optional[QMI_LockTokenDescriptor] = None
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.method_name = method_name
        self.method_args = method_args
        self.method_kwargs = method_kwargs
        self.lock_token = lock_token


class QMI_MethodRpcReplyMessage(QMI_ReplyMessage):
    """Message sent back to an RPC client with the result of a remote method invocation.

    Attributes:
        state: Either `RESULT_IS_VALUE`, `RESULT_IS_EXCEPTION` or `OBJECT_IS_LOCKED`.
        result: Return value from the method or exception raised by the method.
    """
    __slots__ = ("state", "result")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 request_id: str,
                 state: QMI_RpcFutureState,
                 result: Any
                 ) -> None:
        super().__init__(source_address, destination_address, request_id)
        self.state = state
        self.result = result


class QMI_RpcFuture(QMI_MessageHandler):
    """Representation of the future completion of a method invoked via RPC.

    An instance of `QMI_RpcFuture` is created when a call to an RPC method
    is dispatched. The future is completed when the RPC method finishes
    running. This happens asynchronously, because the real RPC method
    execution happens in a background thread.

    This class is used internally when the application invokes a method
    on an RPC proxy object. In this case the proxy method will block until
    the future is completed. Alternatively, when the application invokes
    a non-blocking RPC call, the proxy method returns an instance of this
    class without waiting for the real method call to end.
    """

    def __init__(self,
                 context: "qmi.core.context.QMI_Context",
                 rpc_object_address: QMI_MessageHandlerAddress,
                 lock_token: Optional[QMI_LockTokenDescriptor],
                 ) -> None:
        future_address = context.make_unique_address("$future_")
        super().__init__(future_address)

        self._result = None
        self._context = context  # The context that manages us.
        self._state = QMI_RpcFutureState.NO_RESULT_YET
        self._cv = threading.Condition(threading.Lock())
        self.rpc_object_address = rpc_object_address
        self.lock_token = lock_token

        context.register_message_handler(self)

    def send_method_rpc_request_message(self,
                                        rpc_method_name: str,
                                        rpc_method_args: tuple,
                                        rpc_method_kwargs: dict
                                        ) -> None:
        """Send a request message to the RPC object to invoke the specified method.

        Parameters:
            rpc_method_name: Name of the method to call.
            rpc_method_args: Tuple of positional arguments.
            rpc_method_kwargs: Dictionary of keyword arguments.
        """
        request = QMI_MethodRpcRequestMessage(
            self.address,
            self.rpc_object_address,
            rpc_method_name,
            rpc_method_args,
            rpc_method_kwargs,
            self.lock_token
        )

        try:
            self._context.send_message(request)
        except QMI_MessageDeliveryException as exc:
            self._set_result(QMI_RpcFutureState.RESULT_IS_EXCEPTION, exc)

    def send_lock_rpc_request_message(self, action: QMI_LockRpcAction) -> None:
        request = QMI_LockRpcRequestMessage(self.address, self.rpc_object_address, self.lock_token, action)
        try:
            self._context.send_message(request)
        except QMI_MessageDeliveryException as exc:
            self._set_result(QMI_RpcFutureState.RESULT_IS_EXCEPTION, exc)

    def handle_message(self, message: QMI_Message) -> None:
        """Called when a reply message is received."""

        if isinstance(message, QMI_MethodRpcReplyMessage):
            # Received result from RPC call.
            self._set_result(message.state, message.result)
        elif isinstance(message, QMI_LockRpcReplyMessage):
            # Response to lock request message.
            self._set_result(QMI_RpcFutureState.RESULT_IS_VALUE, message.lock_token)
        elif isinstance(message, QMI_ErrorReplyMessage):
            # Delivery of RPC request failed.
            self._set_result(QMI_RpcFutureState.RESULT_IS_EXCEPTION,
                             QMI_MessageDeliveryException(message.error_msg))
        else:
            _logger.error("Future for %s.%s received unexpected message type %r",
                          self.rpc_object_address.context_id,
                          self.rpc_object_address.object_id,
                          type(message))

    def _set_result(self, state: QMI_RpcFutureState, result: Any) -> None:
        """Store the result received from the RPC reply and wake up any task waiting for this result."""
        if state not in (QMI_RpcFutureState.RESULT_IS_EXCEPTION,
                         QMI_RpcFutureState.RESULT_IS_VALUE,
                         QMI_RpcFutureState.OBJECT_IS_LOCKED):
            _logger.error("Future for %s.%s received unexpected state %r",
                          self.rpc_object_address.context_id,
                          self.rpc_object_address.object_id,
                          state)
            return

        with self._cv:
            if self._state != QMI_RpcFutureState.NO_RESULT_YET:
                _logger.error("Future for %s.%s received duplicate reply message",
                              self.rpc_object_address.context_id,
                              self.rpc_object_address.object_id)
                return

            self._state = state
            self._result = result
            self._cv.notify_all()

    def wait(self, timeout: Optional[float] = None) -> Any:
        """Wait until the RPC call completes.

        Parameters:
            timeout: Maximum wait time in seconds, or None to wait forever.

        Returns:
            The return value from the associated RPC method call.

        Raises:
            QMI_RpcTimeoutException: If the timeout expires before the RPC call completes.
            Exception: If the associated RPC method call raised an exception.
        """
        if timeout is not None:
            time_limit = time.monotonic() + timeout

        try:
            with self._cv:
                while True:
                    # Check state of the future.
                    if self._state == QMI_RpcFutureState.RESULT_IS_EXCEPTION:
                        if not isinstance(self._result, BaseException):
                            raise QMI_RuntimeException("Received invalid exception value from RPC call")
                        raise self._result

                    if self._state == QMI_RpcFutureState.RESULT_IS_VALUE:
                        return self._result
                    elif self._state == QMI_RpcFutureState.OBJECT_IS_LOCKED:
                        raise QMI_RuntimeException("The object is locked by another proxy")
                    else:
                        # No result yet.
                        pass

                    # No result yet, see if we are supposed to wait longer.
                    if timeout is None:
                        wait_result = self._cv.wait(timeout)
                    else:
                        wait_result = self._cv.wait(max(0.0, time_limit - time.monotonic()))

                    if not wait_result:
                        # Timeout expired!
                        raise QMI_RpcTimeoutException("Timeout in RPC call.")

        finally:
            # The QMI_RpcFuture has now become useless.
            self._context.unregister_message_handler(self)


def non_blocking_rpc_method_call(context: "qmi.core.context.QMI_Context",
                                 rpc_object_address: QMI_MessageHandlerAddress,
                                 method_name: str,
                                 rpc_lock_token: Optional[QMI_LockTokenDescriptor],
                                 *args: Any,
                                 **kwargs: Any
                                 ) -> Any:
    """Helper function that performs a non-blocking call to a specific method of the target RPC object."""
    if "rpc_timeout" in kwargs:
        raise RuntimeError("rpc_timeout parameter makes no sense for non-blocking invocation.")

    future = QMI_RpcFuture(context, rpc_object_address, rpc_lock_token)
    future.send_method_rpc_request_message(method_name, args, kwargs)
    return future


def blocking_rpc_method_call(context: "qmi.core.context.QMI_Context",
                             rpc_object_address: QMI_MessageHandlerAddress,
                             method_name: str,
                             rpc_lock_token: Optional[QMI_LockTokenDescriptor],
                             *args: Any,
                             rpc_timeout: Optional[float] = None,
                             **kwargs: Any
                             ) -> Any:
    """Helper function that performs a blocking call to a specific method of the target RPC object."""
    future = QMI_RpcFuture(context, rpc_object_address, rpc_lock_token)
    future.send_method_rpc_request_message(method_name, args, kwargs)
    return future.wait(rpc_timeout)


class QMI_RpcNonBlockingProxy:
    """Proxy class for RPC objects that performs non-blocking calls. Direct instantiation is not recommended.

    This is always also instantiated in `QMI_RpcProxy` as `self.rpc_nonblocking` attribute. Typically, if user wants
    to use a non-blocking call with an RPC object `rpc_proxy`, they would do:
    ```python
    future = rpc_proxy.rpc_nonblocking.some_rpc_command(args)
    # Other stuff can be done here in the meanwhile, and the `proxy` is not blocked while waiting to return
    retval = future.wait()
    ```
    Instead of the usual
    ```python
    retval = rpc_proxy.some_rpc_commands(args)
    ```
    """

    def __init__(self, context: "qmi.core.context.QMI_Context", descriptor: RpcObjectDescriptor) -> None:

        self._context = context
        self._rpc_object_address = descriptor.address
        self._rpc_class_fqn = ".".join((descriptor.interface.rpc_class_module, descriptor.interface.rpc_class_name))
        self._lock_token: Optional[QMI_LockTokenDescriptor] = None

        # Set docstring.
        setattr(self, "__doc__", descriptor.interface.rpc_class_docstring)

        # Helper function used to create a new scope such that each method created in the loop below uses the intended
        # method name.
        def make_rpc_forward_function(method_name: str):
            return lambda self, *args, **kwargs: \
                non_blocking_rpc_method_call(self._context, self._rpc_object_address, method_name,
                                             self._lock_token, *args, **kwargs)

        # Add methods.
        for method_descriptor in descriptor.interface.methods:
            # Generate a function that forward calls to itself to the corresponding RPC method of the peer context.
            method = make_rpc_forward_function(method_descriptor.name)

            # Update special attributes to make the forward function look like the method it is a proxy for.
            docstring = "rpc proxy for {}{} method of {} instance".format(method_descriptor.name,
                                                                          method_descriptor.signature,
                                                                          self._rpc_class_fqn)

            if method_descriptor.docstring:
                docstring = docstring + "\n\n" + method_descriptor.docstring

            setattr(method, "__name__", method_descriptor.name)
            setattr(method, "__qualname__", type(self).__name__ + "." + method_descriptor.name)
            setattr(method, "__doc__", docstring)

            # Invoke the __get__() descriptor function to create a bound method and attach it to the proxy instance
            # as an attribute.
            setattr(self, method_descriptor.name, method.__get__(self))

    def __repr__(self) -> str:
        return "<non-blocking rpc proxy for {} ({})>".format(self._rpc_object_address, self._rpc_class_fqn)


class QMI_RpcProxy:
    """Proxy class for RPC objects that performs blocking calls. All RPC objects created return this proxy class to
    enable RPC communication between objects.

    Direct instantiation of this class is not meant to be done by users; internal use only!
    """

    def __init__(self, context: "qmi.core.context.QMI_Context", descriptor: RpcObjectDescriptor) -> None:

        self._context = context
        self._rpc_object_address = descriptor.address
        self._rpc_class_fqn = ".".join((descriptor.interface.rpc_class_module, descriptor.interface.rpc_class_name))
        self._lock_token: Optional[QMI_LockTokenDescriptor] = None

        # Helper function used to create a new scope such that each method created in the loop below uses the intended
        # method name.
        def make_rpc_forward_function(method_name: str):
            return lambda self, *args, **kwargs: \
                blocking_rpc_method_call(self._context, self._rpc_object_address, method_name, self._lock_token,
                                         *args, **kwargs)

        # Set docstring.
        setattr(self, "__doc__", descriptor.interface.rpc_class_docstring)

        # Add constants.
        for constant_descriptor in descriptor.interface.constants:
            setattr(self, constant_descriptor.name, constant_descriptor.value)

        # Add methods.
        for method_descriptor in descriptor.interface.methods:
            # Generate a function that forward calls to itself to the corresponding RPC method of the peer context.
            method = make_rpc_forward_function(method_descriptor.name)

            # Update special attributes to make the forward function look like the method it is a proxy for.
            docstring = "rpc proxy for {}{} method of {} instance".format(method_descriptor.name,
                                                                          method_descriptor.signature,
                                                                          self._rpc_class_fqn)
            if method_descriptor.docstring:
                docstring = docstring + "\n\n" + method_descriptor.docstring

            setattr(method, "__name__", method_descriptor.name)
            setattr(method, "__qualname__", type(self).__name__ + "." + method_descriptor.name)
            setattr(method, "__doc__", docstring)

            # Invoke the __get__() descriptor function to create a bound method and attach it to the proxy instance
            # as an attribute.
            setattr(self, method_descriptor.name, method.__get__(self))

        # Add signals.
        for signal_descriptor in descriptor.interface.signals:
            subscriber = QMI_SignalSubscriber(
                self._context,
                self._rpc_object_address.context_id,
                self._rpc_object_address.object_id,
                signal_descriptor.name,
                signal_descriptor.arg_types
            )
            setattr(self, signal_descriptor.name, subscriber)

        # Add non-blocking proxy.
        self.rpc_nonblocking = QMI_RpcNonBlockingProxy(context, descriptor)

    def __enter__(self) -> "QMI_RpcProxy":
        """The context manager definition is needed for the proxy as it will always be returned from QMI contexts,
        instead of the actual RPC object instance. Trying to use context management directly on this class will
        cause recursion error, but for actual RPC objects not.

        The with keyword checks if the `QMI_RpcProxy` class (not instance) implements the context manager protocol,
        i.e. `__enter__` and `__exit__`, so they exist here as stubs. These stubs make an RPC to the actual `__enter__`
        and `__exit__` methods on the relevant RPC object. If in those classes `__enter__` and `__exit__` are
        decorated as `rpc_methods`, we do not see a recursion error. See for further details in:
        https://docs.python.org/3/reference/datamodel.html#special-method-lookup

        `rpc_method` decorated `__enter__` and `__exit__` methods are currently implemented in `QMI_Instrument` and
        `QMI_TaskRunner` classes.
        """
        self.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.__exit__()
        return

    def __repr__(self) -> str:
        return "<rpc proxy for {} ({})>".format(self._rpc_object_address, self._rpc_class_fqn)

    @property
    def address(self) -> str:
        """Return the "address" of the proxy with context name and object name.

        Returns:
            String "{context name}.{name}".
        """
        return str(self._rpc_object_address)

    def lock(self, timeout: float = 0.0, lock_token: Optional[str] = None) -> bool:
        """Substitutes the `lock` method stub of QMI_RpcObject."""
        # Create a lock token for this proxy and try to lock the object.
        their_lock_token = "None"
        if lock_token is not None:
            my_lock_token = QMI_LockTokenDescriptor(self._context.name, lock_token)

        else:
            my_lock_token = self._context.make_unique_token(prefix="$lock_")

        if timeout > 0.0:
            hardcoded_period = 0.1  # minimum time period for tries within timeout: timeout < 0.1 is practically 0.1.
            time_start = time.monotonic()
            while (time.monotonic() - time_start) < timeout:
                loop_start = time.monotonic()
                future = QMI_RpcFuture(self._context, self._rpc_object_address, my_lock_token)
                future.send_lock_rpc_request_message(QMI_LockRpcAction.ACQUIRE)

                # If the lock is granted, the returned lock token matches our lock token.
                their_lock_token = future.wait()
                if their_lock_token == my_lock_token:
                    _logger.debug("%s locked with %s", self._rpc_object_address, my_lock_token)
                    self._lock_token = my_lock_token
                    self.rpc_nonblocking._lock_token = my_lock_token
                    return True

                loop_end = time.monotonic()
                if (loop_end - loop_start) < hardcoded_period:
                    time.sleep(hardcoded_period - (loop_end - loop_start))

        else:
            # Just try once
            future = QMI_RpcFuture(self._context, self._rpc_object_address, my_lock_token)
            future.send_lock_rpc_request_message(QMI_LockRpcAction.ACQUIRE)

            # If the lock is granted, the returned lock token matches our lock token.
            their_lock_token = future.wait()
            if their_lock_token == my_lock_token:
                _logger.debug("%s locked with %s", self._rpc_object_address, my_lock_token)
                self._lock_token = my_lock_token
                self.rpc_nonblocking._lock_token = my_lock_token
                return True

        _logger.debug("%s lock denied, already locked with %s", self._rpc_object_address, their_lock_token)
        return False

    def unlock(self, lock_token: Optional[str] = None) -> bool:
        """Substitutes the `unlock` method stub of QMI_RpcObject."""
        if lock_token is not None:
            # Try to unlock with the custom lock token.
            my_lock_token = QMI_LockTokenDescriptor(self._context.name, lock_token)
            future = QMI_RpcFuture(self._context, self._rpc_object_address, my_lock_token)

        else:
            # Try to unlock the object with this proxy's lock token.
            future = QMI_RpcFuture(self._context, self._rpc_object_address, self._lock_token)

        future.send_lock_rpc_request_message(QMI_LockRpcAction.RELEASE)

        # If the token was accepted, the returned lock token matches our lock token (or theirs is None if the object
        # was not even locked).
        their_lock_token = future.wait()
        if their_lock_token is None:
            _logger.debug("%s unlocked with %s", self._rpc_object_address, self._lock_token)
            self._lock_token = None  # do not reuse tokens
            self.rpc_nonblocking._lock_token = None
            return True
        else:
            _logger.debug("%s unlock with %s denied, locked with %s", self._rpc_object_address, self._lock_token,
                          their_lock_token)
            return False

    def force_unlock(self) -> None:
        """Substitutes the `force_unlock` method stub of QMI_RpcObject."""
        # Send force-release message.
        future = QMI_RpcFuture(self._context, self._rpc_object_address, self._lock_token)
        future.send_lock_rpc_request_message(QMI_LockRpcAction.FORCE_RELEASE)

        # Check that the lock was removed (token is None).
        their_lock_token = future.wait()
        if their_lock_token is None:
            _logger.debug("%s unlocked forcefully", self._rpc_object_address)
            self._lock_token = None
            self.rpc_nonblocking._lock_token = None
        else:
            _logger.warning("%s force unlock failed!", self._rpc_object_address)

    def is_locked(self) -> bool:
        """Substitutes the `is_locked` method stub of QMI_RpcObject."""
        # Send query.
        future = QMI_RpcFuture(self._context, self._rpc_object_address, self._lock_token)
        future.send_lock_rpc_request_message(QMI_LockRpcAction.QUERY)

        # If the object is lock, it will have a lock token set.
        lock_token = future.wait()
        return lock_token is not None


def rpc_method(method: _T) -> _T:
    """Decorator to indicate that a method can be called via RPC."""
    method._rpc_method = True  # type: ignore
    return method


def is_rpc_method(object: Any) -> bool:
    """Return True if the object is a method that can be called via RPC."""
    return inspect.isfunction(object) and getattr(object, "_rpc_method", False)


class _RpcObjectMetaClass(ABCMeta):
    """Meta-class used to create `QMI_RpcObject` and its subclasses.

    This meta-class extracts a list of signals published by the class.
    The list of signals is inserted as a class attribute named `_qmi_signals`.
    """

    def __init__(cls, name: str, bases: tuple, dct: dict) -> None:

        # Let "type" do its thing.
        super().__init__(name, bases, dct)

        # The original class must not yet have a _qmi_signals attribute.
        assert "_qmi_signals" not in dct

        # Scan the newly created QMI_RpcObject subclass for class attributes of type QMI_Signal.
        # These are markers to announce signals that the class may publish.
        signals = []
        for attr_name, attr_value in inspect.getmembers(cls, lambda member: isinstance(member, QMI_Signal)):
            if not is_valid_object_name(attr_name):
                raise QMI_UsageException("Invalid signal name {!r} in class {}".format(attr_name, cls.__name__))

            signals.append(SignalDescription(name=attr_name, arg_types=attr_value.arg_types))

        # Add the list of signals as an attribute of the newly created QMI_RpcObject subclass.
        cls._qmi_signals = tuple(signals)


class QMI_RpcObject(metaclass=_RpcObjectMetaClass):
    """Base class for classes which expose (a subset of) their methods via RPC.

    Each instance of `QMI_RpcObject` has a name, unique within its context.
    The context maintains a list of `QMI_RpcObject` instances.

    Subclasses of `QMI_RpcObject` apply the `@rpc_method` decorator to
    (a subset of) their methods to mark them as callable via RPC.

    Subclasses of `QMI_RpcObject` may choose to export (a subset of)
    their constant class attributes to be accessible directly via the proxy.
    This is done by creating a class attribute `_rpc_constants` holding
    a list of attribute names to be exported.

    Each instance of `QMI_RpcObject` runs in a separate thread. It is not allowed
    to invoke methods of the `QMI_RpcObject` directly from outside the class.
    Instead, the proper way to access an instance of `QMI_RpcObject` is to
    invoke a method on a special "proxy" object, which translates the call
    into an RPC request, which triggers an invocation of the real method of
    the `QMI_RpcObject` instance.

    Instances of `QMI_RpcObject` may publish QMI signals. Each instance may
    register a set of signals. Once registered, such signals can be published
    into the QMI network and routed to subscribed receivers.
    """

    @classmethod
    def get_category(cls) -> Optional[str]:
        """Return the optional name of the category this object belongs to.

        A category name is a free-form string that has no special significance.
        Its purpose is to distinguish between groups of RPC objects that fulfill
        similar roles.
        """
        return None

    def __init__(self,
                 context: 'qmi.core.context.QMI_Context',
                 name: str,
                 signal_declaration_class: Optional[type] = None
                 ) -> None:
        """Initialize the object.

        Instances of QMI_RpcObject are created and managed by the context.
        They should not normally be instantiated directly by the application.

        Parameters:
            context: Instance of `QMI_Context` that will manage this object.
            name: Unique name of this object instance.
            signal_declaration_class: Optional separate class which declares
                the signals published by this object.
        """
        self._context = context
        self._name = name

        # Create an RpcObjectDescriptor for this QMI_RpcObject instance.
        #
        # In the general case, the "signal_declaration_class" is identical to
        # the "rpc_object_class". This simply means that the class which
        # implements the RPC methods, is also the class which declares signals.
        # As an exception to this rule, QMI_TaskRunner will specify a different
        # signal declaration class.
        if signal_declaration_class is None:
            signal_declaration_class = type(self)

        self.rpc_object_descriptor = RpcObjectDescriptor(
            address=QMI_MessageHandlerAddress(context_id=context.name, object_id=name),
            category=self.get_category(),
            interface=make_interface_descriptor(type(self), signal_declaration_class)
        )

        # Create instances of QMI_RegisteredSignal for each signal type that this class may publish.
        # Insert these QMI_RegisteredSignal instances as attributes of the RpcObject instance.
        #
        # Note: The class attribute "_qmi_signals" will be created by the metaclass.
        declared_signals = self._qmi_signals  # type: ignore  # pylint: disable=no-member
        for sig_desc in declared_signals:
            sig = QMI_RegisteredSignal(context, name, sig_desc.name, sig_desc.arg_types)
            setattr(self, sig_desc.name, sig)

    def __repr__(self) -> str:
        return "{}({!r})".format(type(self).__name__, self._name)

    @rpc_method
    def __enter__(self):
        raise NotImplementedError(f"{type(self)} is not meant to be used with context manager.")

    def lock(self, timeout: float = 0.0, lock_token: Optional[str] = None) -> bool:
        """Lock the remote object. If timeout is given, try every 0.1s within the given timeout value. The remote
        object can be locked with an optional custom lock token by giving a string into `lock_token` keyword argument.

        If successful, this proxy is the only proxy that can invoke RPC methods on the remote object; other proxies
        will receive an "object is locked" response. The return value indicates if the lock was granted; a denied lock
        means the object was already locked by another proxy.

        Do not override this stub method in subclasses. It has already been implemented in QMI_RpcProxy.
        """
        raise NotImplementedError("QMI_RpcObject.lock not implemented")

    def unlock(self, lock_token: Optional[str] = None) -> bool:
        """Unlock the remote object.

        Without optional parameters, this is only allowed by the proxy that initially locked the object. By giving
        the lock token as an input parameter, the specific object locked by this token can be unlocked.
        The return value indicates if the unlocking was successful.

        Do not override this stub method in subclasses. It has already been implemented in QMI_RpcProxy.
        """
        raise NotImplementedError("QMI_RpcObject.unlock not implemented")

    def is_locked(self) -> bool:
        """Query if the remote object is locked.

        Do not override this stub method in subclasses. It has already been implemented in QMI_RpcProxy.
        """
        raise NotImplementedError("QMI_RpcObject.is_locked not implemented")

    def force_unlock(self) -> None:
        """Forcefully unlock the remote object.

        This unlocks the object, regardless of who owns the lock. This allows you to unlock an object if the locking
        proxy has been destroyed without unlocking.

        Use this with care.

        Do not override this stub method in subclasses. It has already been implemented in QMI_RpcProxy.
        """
        raise NotImplementedError("QMI_RpcObject.force_unlock not implemented")

    def release_rpc_object(self) -> None:
        """This method is called just before the RPC object is removed from the context.

        When an RPC object is removed from the context (via context.remove_rpc_object()),
        this method is invoked to release any resources used by the RPC object instance.

        Subclasses may override this method to release specific resources.
        The default implementation does nothing.
        """

    @rpc_method
    def get_name(self) -> str:
        """Return the name of this object.

        Returns:
            name attribute.
        """
        return self._name

    @rpc_method
    def get_signals(self) -> List[SignalDescription]:
        """Return a list of signals that can be published by this object.

        Returns:
            List consisting of qmi_signals attributes.
        """

        # Note: The class attribute "_qmi_signals" will be created by the metaclass.
        return list(self._qmi_signals)  # type: ignore  # pylint: disable=no-member


def make_interface_descriptor(rpc_object_class: Type[QMI_RpcObject],
                              signal_declaration_class: Optional[Type[QMI_RpcObject]] = None
                              ) -> RpcInterfaceDescriptor:
    """Create a description of the (subset of the) interface of the specified
    `QMI_RpcObject` subclass that can be accessed via RPC. Signal declarations
    are taken from the specified signal declaration class, which may be a
    different class than that from which the RPC methods are extracted.
    """

    # Use the RPC object class as the class to extract signal declarations from if no signal declaration class was
    # provided by the caller.
    if signal_declaration_class is None:
        signal_declaration_class = rpc_object_class

    # Extract RPC method declarations.
    methods = []
    for name, member in inspect.getmembers(rpc_object_class, is_rpc_method):
        if name in ("lock", "unlock", "force_unlock", "is_locked"):
            raise QMI_UsageException(f"`{name}` is a protected method name")

        signature = str(inspect.signature(member))
        docstring = member.__doc__
        methods.append(RpcMethodDescriptor(name, signature, docstring))

    # Extract signal declarations.
    signals = []
    for signal_description in signal_declaration_class._qmi_signals:
        name = signal_description.name
        arg_types = "(" + ", ".join(arg_type.__name__ for arg_type in signal_description.arg_types) + ")"
        signals.append(RpcSignalDescriptor(name, arg_types))

    # Extract constant declarations.
    constant_names = set()
    for base in inspect.getmro(rpc_object_class):
        if hasattr(base, "_rpc_constants"):
            constant_names.update(getattr(base, "_rpc_constants"))

    # Extract constant values.
    constants = []
    for constant_name in constant_names:
        assert hasattr(rpc_object_class, constant_name)
        constant_value = getattr(rpc_object_class, constant_name)
        assert not inspect.isfunction(constant_value)
        constants.append(RpcConstantDescriptor(constant_name, constant_value))

    # Create interface descriptor.
    return RpcInterfaceDescriptor(rpc_object_class.__module__, rpc_object_class.__name__, rpc_object_class.__doc__,
                                  constants, methods, signals)


class _RpcThread(QMI_Thread):
    """Dedicated thread for executing methods invoked via RPC.

    An instance of this class handles RPC invocations for a single instance
    of `QMI_RpcObject`. A separate instance of `RpcThread` is created for each
    RPC object instance.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.
    """

    def __init__(self,
                 context: 'qmi.core.context.QMI_Context',
                 rpc_object_maker: Callable[[], QMI_RpcObject]
                 ) -> None:
        super().__init__()
        self._context = context  # We need to know the context, to send replies to RPC requests.
        self._rpc_object_maker = rpc_object_maker
        self._locking_token: Optional[QMI_LockTokenDescriptor] = None  # thread maintains lock token
        self._cv = threading.Condition(threading.Lock())
        self._fifo: deque = deque()
        self._rpc_object: Optional[QMI_RpcObject] = None
        self._exception:  Optional[BaseException] = None

    def _handle_lock_rpc_request(self, request: QMI_LockRpcRequestMessage) -> QMI_LockRpcReplyMessage:
        """Handle a lock message."""
        assert self._rpc_object is not None

        return_token: Optional[QMI_LockTokenDescriptor]
        if request.lock_action == QMI_LockRpcAction.ACQUIRE and request.lock_token is not None:
            # Lock request; check if it is allowed.
            if self._locking_token is None:
                # Object was not locked, lock it by storing the provided token.
                self._locking_token = return_token = request.lock_token
                _logger.info("%s locked with %s!", self._rpc_object.get_name(), request.lock_token)
            elif self._locking_token != request.lock_token:
                # Object was already locked and lock token does not match.
                return_token = QMI_LockTokenDescriptor(self._context.name, ACCESS_DENIED_TOKEN_PLACEHOLDER)
                _logger.warning("Lock request (token=%s) for %s failed! Already locked with token=%s.",
                                request.lock_token,
                                self._rpc_object.get_name(),
                                self._locking_token)
            else:
                # Lock token matches, but object is already locked, nothing to do.
                return_token = self._locking_token

        elif request.lock_action == QMI_LockRpcAction.RELEASE:
            # Lock release request; check if it is allowed.
            if self._locking_token is None:
                # Object was not locked, nothing to do.
                return_token = self._locking_token
            elif self._locking_token == request.lock_token:
                # Lock token matches, unlock by clearing the locking token.
                self._locking_token = return_token = None
                _logger.info("%s unlocked with %s!", self._rpc_object.get_name(), request.lock_token)
            else:
                # Lock token does not match.
                return_token = QMI_LockTokenDescriptor(self._context.name, ACCESS_DENIED_TOKEN_PLACEHOLDER)
                _logger.warning("Unlocking request (token=%s) for %s failed! Locked with token=%s.",
                                request.lock_token,
                                self._rpc_object.get_name(),
                                self._locking_token)

        elif request.lock_action == QMI_LockRpcAction.FORCE_RELEASE:
            # Force release of lock irrespective of requesting proxy.
            if self._locking_token is not None:
                self._locking_token = return_token = None
                _logger.warning("%s forcefully unlocked!", self._rpc_object.get_name())

        elif request.lock_action == QMI_LockRpcAction.QUERY:
            # Nothing to do here; reply will contain the locking token (if any), is_locked() method of proxy
            # will check and return appropriate boolean value.
            if self._locking_token is None:
                return_token = None
            else:
                return_token = QMI_LockTokenDescriptor(self._context.name, OBJECT_LOCKED_TOKEN_PLACEHOLDER)

        else:
            raise ValueError(f"Unknown lock action: {request.lock_action!r}")

        reply = QMI_LockRpcReplyMessage(
            source_address=request.destination_address,
            destination_address=request.source_address,
            request_id=request.request_id,
            lock_token=return_token
        )
        return reply

    def _handle_method_rpc_request(self, request: QMI_MethodRpcRequestMessage) -> QMI_MethodRpcReplyMessage:
        """Handle RPC method request."""
        assert self._rpc_object is not None

        # RPC method call - need to check if the caller may invoke the RPC method: allowed if the object is not
        # locked (token is None) or if the provided lock token matches the locking token.
        if self._locking_token is None or self._locking_token == request.lock_token:
            # Invoke the method; this can raise an exception or return a result.
            try:
                method = self._check_and_get_method(request)
                result_type = QMI_RpcFutureState.RESULT_IS_VALUE
                result = method(*request.method_args, **request.method_kwargs)

            except BaseException as exception:
                _logger.debug("RPC method call failed", exc_info=True)
                result_type = QMI_RpcFutureState.RESULT_IS_EXCEPTION
                result = exception
        else:
            _logger.error("%s locked, method request without lock token is denied", self._rpc_object._name)
            result_type = QMI_RpcFutureState.OBJECT_IS_LOCKED
            result = None

        reply = QMI_MethodRpcReplyMessage(
            source_address=request.destination_address,
            destination_address=request.source_address,
            request_id=request.request_id,
            state=result_type,
            result=result
        )
        return reply

    def _check_and_get_method(self, request: QMI_MethodRpcRequestMessage):
        """Check if the object has the method requested and is RPC callable; if so, return it."""
        assert self._rpc_object is not None

        # Check that the method exists.
        if not hasattr(self._rpc_object, request.method_name):
            raise QMI_UnknownRpcException("Object {} of type {} does not have method {}"
                                          .format(request.destination_address.object_id,
                                                  type(self._rpc_object).__name__,
                                                  request.method_name))

        # Check that the method was marked as RPC-callable.
        method = getattr(self._rpc_object, request.method_name)
        is_rpc_callable = getattr(method, "_rpc_method", False)
        if not is_rpc_callable:
            raise QMI_UnknownRpcException("Method {!r} is not RPC-callable!".format(request.method_name))

        return method

    def _reject_remaining_requests(self) -> None:
        """Reject any requests that are still in our queue when the RPC object shuts down."""

        while True:
            # Get next pending request.
            with self._cv:
                if not self._fifo:
                    break
                request = self._fifo.popleft()

            # Sanity check (this has already been checked by the RpcObjectManager).
            assert isinstance(request, (QMI_MethodRpcRequestMessage, QMI_LockRpcRequestMessage))

            # Send error reply for this request.
            reply = QMI_ErrorReplyMessage(source_address=request.destination_address,
                                          destination_address=request.source_address,
                                          request_id=request.request_id,
                                          error_msg="")
            try:
                self._context.send_message(reply)
            except QMI_MessageDeliveryException:
                # Ignore errors while sending the error reply.
                _logger.debug("Failed to send RPC error reply to %s.%s",
                              request.source_address.context_id,
                              request.source_address.object_id)

    def _request_shutdown(self) -> None:
        # Notify the thread so that it can end its request loop.
        assert self._shutdown_requested
        with self._cv:
            self._cv.notify_all()

    def rpc_object(self) -> QMI_RpcObject:
        """Return the actual RpcObject instance managed by this thread.

        Wait until initialization of the RpcObject instance is finished,
        if necessary. If initialization was successful, return the object
        instance. Otherwise, re-raise the exception that occurred during
        initialization.
        """
        with self._cv:
            while (self._rpc_object is None) and (self._exception is None):
                self._cv.wait()
            if self._exception is not None:
                raise self._exception

            assert isinstance(self._rpc_object, QMI_RpcObject)
            return self._rpc_object

    def run(self) -> None:
        """This method runs inside the RPC thread."""
        _logger.debug("Starting RPC thread")

        assert self._rpc_object is None
        assert self._exception is None

        try:
            # Construct and initialize RpcObject instance.
            rpc_object = self._rpc_object_maker()
            if not isinstance(rpc_object, QMI_RpcObject):
                raise TypeError("Expecting QMI_RpcObject but got {}".format(type(rpc_object)))
        except BaseException as exception:
            # Initialization failed. Store the exception.
            _logger.warning("Initialization of RpcObject failed", exc_info=True)
            with self._cv:
                self._exception = exception
                self._cv.notify_all()

            # Stop thread.
            _logger.info("Stopping RPC thread (initialization failed)")
            return

        # Notify the outside world that initialization is finished.
        with self._cv:
            self._rpc_object = rpc_object
            self._cv.notify_all()

        # Request handling phase.
        while True:
            with self._cv:
                # Wait for RPC request or shutdown request.
                while (not self._shutdown_requested) and (not self._fifo):
                    self._cv.wait()

                # End the request loop if shutdown requested.
                if self._shutdown_requested:
                    break

                request = self._fifo.popleft()

            # Process request.
            reply: Union[QMI_MethodRpcReplyMessage, QMI_LockRpcReplyMessage]
            if isinstance(request, QMI_MethodRpcRequestMessage):
                reply = self._handle_method_rpc_request(request)
            elif isinstance(request, QMI_LockRpcRequestMessage):
                reply = self._handle_lock_rpc_request(request)
            else:
                raise ValueError("Unknown request type: {}".format(type(request)))

            # Send reply.
            try:
                self._context.send_message(reply)
            except QMI_MessageDeliveryException:
                # Catch exceptions from sending message (avoid crashing the RPC thread on message delivery error).
                _logger.error(
                    "Failed to send RPC reply message from %s.%s to %s.%s",
                    request.destination_address.context_id,
                    request.destination_address.object_id,
                    request.source_address.context_id,
                    request.source_address.object_id
                )

            del request, reply

        # Reject any requests that are still in our queue.
        self._reject_remaining_requests()

        # Tell RPC object to release resources.
        try:
            rpc_object.release_rpc_object()
        except BaseException:
            # Log exceptions during resource release.
            _logger.exception("Failed to release RPC object")

        _logger.debug("Stopping RPC thread")

    def push_rpc_request(self, rpc_request: Union[QMI_MethodRpcRequestMessage, QMI_LockRpcRequestMessage]) -> None:
        """Push an RPC request into the request queue and notify the thread."""
        with self._cv:
            self._fifo.append(rpc_request)
            self._cv.notify_all()


class RpcObjectManager(QMI_MessageHandler):
    """Manages a single instance of QMI_RpcObject.

    A dedicated `RpcObjectManager` is created by the context for each instance
    of `QMI_RpcObject`. The `RpcObjectManager` owns a single `RpcThread` instance
    (which, in turn, owns a single `QMI_RpcObject` instance).

    An `RpcObjectManager` receives RPC request messages on behalf of the
    RPC object. It pushes these messages into a queue, from where they are
    picked up and handled by the RpcThread.

    The message handler address of the RpcObjectManager is based on
    the name of the associated RPC object.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.
    """

    def __init__(self,
                 address: QMI_MessageHandlerAddress,
                 context: 'qmi.core.context.QMI_Context',
                 rpc_object_maker: Callable[[], QMI_RpcObject]
                 ) -> None:
        """Initialize the RPC object manager.

        Parameters:
            address: Address of the RPC object.
                This instance of `RpcObjectManager` will be registered as
                message handler for this address.
            context: QMI context in which this RPC object will exist.
            rpc_object_maker: Function which creates the actual RPC object instance.
        """
        super().__init__(address)
        self._context = context
        self._rpc_thread: Optional[_RpcThread] = None
        self._rpc_object_maker = rpc_object_maker
        self._stop_lock = threading.Lock()
        self._running = False

    def start(self) -> None:
        """Create a background thread and start handling RPC calls."""
        assert self._rpc_thread is None
        self._rpc_thread = _RpcThread(self._context, self._rpc_object_maker)
        self._rpc_thread.start()
        self._running = True

    def stop(self) -> None:
        """Stop the background thread and clean it up."""
        assert self._rpc_thread is not None
        with self._stop_lock:
            self._running = False
        self._rpc_thread.shutdown()
        self._rpc_thread.join()
        self._rpc_thread = None

    def make_proxy(self) -> QMI_RpcProxy:
        """Return a new instance of the RpcProxy class corresponding to the RPC object.

        Wait until initialization of the RPC object instance is finished,
        if necessary. If initialization was successful, create and return
        a RpcProxy instance for the RPC object.

        Otherwise re-raise the exception that occurred during initialization.
        """

        assert self._rpc_thread is not None

        # This may raise an exception!
        rpc_object = self._rpc_thread.rpc_object()
        return QMI_RpcProxy(self._context, rpc_object.rpc_object_descriptor)

    def handle_message(self, message: QMI_Message) -> None:
        """Called when a QMI message is delivered for our RPC object."""

        if not isinstance(message, (QMI_MethodRpcRequestMessage, QMI_LockRpcRequestMessage)):
            _logger.error("Received unknown message type %r from %s.%s",
                          type(message),
                          message.source_address.context_id,
                          message.source_address.object_id)
            return

        with self._stop_lock:
            # Reject message if the object is already stopped (or stopping).
            if not self._running:
                raise QMI_MessageDeliveryException("RPC object {}.{} already stopped"
                                                   .format(self.address.context_id, self.address.object_id))

            # The thread is still running, so we can safely push the message to the thread.
            assert self._rpc_thread is not None
            self._rpc_thread.push_rpc_request(message)

    def rpc_object(self) -> QMI_RpcObject:
        """Return the RPC object instanced managed by this `RpcObjectManager`."""
        assert self._rpc_thread is not None
        return self._rpc_thread.rpc_object()


# Imports needed only for static typing.
if typing.TYPE_CHECKING:
    import qmi.core.context



"""Unit test for Tektronix AFG31000 driver."""

import math
from typing import cast
import unittest
from unittest.mock import MagicMock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.tektronix import Waveform, BurstMode, TriggerEdge, Tektronix_Afg31000


class TestAFG31000(unittest.TestCase):

    def setUp(self):
        qmi.start("TestContext")
        self._transport_mock = MagicMock(spec=QMI_TcpTransport)
        with patch(
                'qmi.instruments.tektronix.afg31000.create_transport',
                return_value=self._transport_mock):
            self.instr: Tektronix_Afg31000 = qmi.make_instrument("instr", Tektronix_Afg31000, "transport_descriptor")
            self.instr = cast(Tektronix_Afg31000, self.instr)

    def tearDown(self):
        qmi.stop()

    def _open_helper(self):
        """Open instrument and check transport calls."""
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.write.assert_called_once_with(b"*CLS\n")
        self._transport_mock.open.reset_mock()
        self._transport_mock.write.reset_mock()

    def _prep_cmd_no_error(self):
        """Prepare transport mock for error check without error."""
        self._transport_mock.read_until.side_effect = [b"1\n", b'0,"No error"\n']

    def _assert_cmd_no_error(self, expect_commands):
        """Assert that the driver sent the specified command."""
        expect_calls = [call(cmd) for cmd in expect_commands]
        expect_calls += [
            call(b"*OPC?\n"),
            call(b"SYST:ERR?\n")
        ]
        self.assertEqual(self._transport_mock.write.mock_calls, expect_calls)

    def _prep_cmd_with_error(self):
        """Prepare transport mock for error check that detects an error."""
        self._transport_mock.read_until.side_effect = [b"1\n", b'-10,"Some error"\n']

    def _assert_cmd_with_error(self, expect_commands):
        """Assert that the driver sent the specified command."""
        expect_calls = [call(cmd) for cmd in expect_commands]
        expect_calls += [
            call(b"*OPC?\n"),
            call(b"SYST:ERR?\n"),
            call(b"*CLS\n")
        ]
        self.assertEqual(self._transport_mock.write.mock_calls, expect_calls)

    def test_open_close(self):
        self._open_helper()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_open_comm_error(self):
        self._transport_mock.write.side_effect = OSError("fake error")
        with self.assertRaises(OSError):
            self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.write.assert_called_once_with(b"*CLS\n")
        self._transport_mock.close.assert_called_once_with()

    def test_get_idn(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"vendor,model,serial,version\n"
        idn = self.instr.get_idn()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"*IDN?\n")
        self.assertEqual(idn.vendor, "vendor")
        self.assertEqual(idn.model, "model")
        self.assertEqual(idn.serial, "serial")
        self.assertEqual(idn.version, "version")

    def test_get_idn_fail(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"nonsense\n"
        with self.assertRaises(QMI_InstrumentException):
            idn = self.instr.get_idn()
        self.instr.close()

    def test_reset(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.reset()
        self.instr.close()
        self._assert_cmd_no_error([b"*CLS\n", b"*RST\n"])

    def test_get_external_reference_enabled(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"INT\n"
        ret = self.instr.get_external_reference_enabled()
        self._transport_mock.write.assert_called_once_with(b"SOURCE:ROSCILLATOR:SOURCE?\n")
        self._transport_mock.write.reset_mock()
        self.assertFalse(ret)

        self._transport_mock.read_until.return_value = b"EXT\n"
        ret = self.instr.get_external_reference_enabled()
        self.instr.close()
        self.assertTrue(ret)

    def test_get_external_reference_enabled_fail(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"ehm\n"
        with self.assertRaises(QMI_InstrumentException):
            ret = self.instr.get_external_reference_enabled()
        self.instr.close()

    def test_set_external_reference_enabled(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_external_reference_enabled(False)
        self._assert_cmd_no_error([b"SOURCE:ROSCILLATOR:SOURCE INT\n"])
        self._transport_mock.write.reset_mock()

        self._prep_cmd_no_error()
        self.instr.set_external_reference_enabled(True)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE:ROSCILLATOR:SOURCE EXT\n"])

    def test_get_output_enabled(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"0\n"
        ret = self.instr.get_output_enabled(1)
        self._transport_mock.write.assert_called_once_with(b"OUTPUT1:STATE?\n")
        self._transport_mock.write.reset_mock()
        self.assertFalse(ret)

        self._transport_mock.read_until.return_value = b"1\n"
        ret = self.instr.get_output_enabled(2)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"OUTPUT2:STATE?\n")
        self.assertTrue(ret)

    def test_get_output_enabled_bad_response(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"ehm\n"
        with self.assertRaises(QMI_InstrumentException):
            ret = self.instr.get_output_enabled(1)
        self.instr.close()

    def test_set_output_enabled(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_output_enabled(1, True)
        self.instr.close()
        self._assert_cmd_no_error([b"OUTPUT1:STATE 1\n"])

    def test_get_waveform(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"SIN\n"
        ret = self.instr.get_waveform(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:FUNCTION:SHAPE?\n")
        self.assertEqual(ret, Waveform.SINE)

    def test_set_waveform(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_waveform(1, Waveform.PULSE)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:FUNCTION:SHAPE PULS\n"])

    def test_get_sweep_enabled(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"CW\n"
        ret = self.instr.get_sweep_enabled(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:FREQUENCY:MODE?\n")
        self.assertFalse(ret)

    def test_get_continuous_frequency(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"1000000\n"
        ret = self.instr.get_continuous_frequency(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:FREQUENCY:CW?\n")
        self.assertEqual(ret, 1.0e6)

    def test_get_continuous_frequency_bad_response(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"big\n"
        with self.assertRaises(QMI_InstrumentException):
            ret = self.instr.get_continuous_frequency(1)
        self.instr.close()

    def test_set_continuous_frequency(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_continuous_frequency(1, 100.0e3)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:FREQUENCY:MODE CW\n",
                                   b"SOURCE1:FREQUENCY:CW 100000.000000\n"])

    def test_get_amplitude(self):
        self._open_helper()
        self._transport_mock.read_until.side_effect = [b"0.5\n", b"3.3\n"]
        ret = self.instr.get_amplitude(1)
        self.instr.close()
        self.assertEqual(self._transport_mock.write.mock_calls, [
            call(b"SOURCE1:VOLTAGE:LEVEL:LOW?\n"),
            call(b"SOURCE1:VOLTAGE:LEVEL:HIGH?\n")
        ])
        self.assertEqual(ret, (0.5, 3.3))

    def test_set_amplitude(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_amplitude(1, 0.0, 5.0)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:VOLTAGE:LEVEL:LOW 0.0000\n",
                                   b"SOURCE1:VOLTAGE:LEVEL:HIGH 5.0000\n"])

    def test_get_pulse_duty_cycle(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"0.15\n"
        ret = self.instr.get_pulse_duty_cycle(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:PULSE:DCYCLE?\n")
        self.assertEqual(ret, 0.15)

    def test_set_pulse_duty_cycle(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_pulse_duty_cycle(1, 0.5)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:PULSE:HOLD DUTY\n",
                                   b"SOURCE1:PULSE:DCYCLE 0.500000\n"])

    def test_get_pulse_delay(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"1E-9\n"
        ret = self.instr.get_pulse_delay(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:PULSE:DELAY?\n")
        self.assertEqual(ret, 1.0e-9)

    def test_set_pulse_delay(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_pulse_delay(1, 0.25e-9)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:PULSE:DELAY 0.00000000025\n"])

    def test_get_burst_mode(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"GAT\n"
        ret = self.instr.get_burst_mode(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:BURST:MODE?\n")
        self.assertEqual(ret, BurstMode.GATED)

    def test_set_burst_mode(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_burst_mode(1, BurstMode.TRIGGERED)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:BURST:MODE TRIG\n"])

    def test_get_burst_count(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"5\n"
        ret = self.instr.get_burst_count(1)
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:BURST:NCYCLES?\n")
        self._transport_mock.write.reset_mock()
        self.assertEqual(ret, 5)

        self._transport_mock.read_until.return_value = b"9.9E+37\n"
        ret = self.instr.get_burst_count(2)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE2:BURST:NCYCLES?\n")
        self.assertEqual(ret, math.inf)

    def test_set_burst_count(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_burst_count(1, 100)
        self._assert_cmd_no_error([b"SOURCE1:BURST:NCYCLES 100\n"])
        self._transport_mock.write.reset_mock()

        self._prep_cmd_no_error()
        self.instr.set_burst_count(1, math.inf)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:BURST:NCYCLES INF\n"])

    def test_get_burst_enabled(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"1\n"
        ret = self.instr.get_burst_enabled(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"SOURCE1:BURST:STATE?\n")
        self.assertTrue(ret)

    def test_set_burst_enabled(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_burst_enabled(1, True)
        self.instr.close()
        self._assert_cmd_no_error([b"SOURCE1:BURST:STATE 1\n"])

    def test_get_external_trigger_enabled(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"EXT\n"
        ret = self.instr.get_external_trigger_enabled()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"TRIGGER:SOURCE?\n")
        self.assertTrue(ret)

    def test_set_external_trigger_enabled(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_external_trigger_enabled(True)
        self.instr.close()
        self._assert_cmd_no_error([b"TRIGGER:SOURCE EXT\n"])

    def test_get_trigger_edge(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"NEG\n"
        ret = self.instr.get_trigger_edge()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"TRIGGER:SLOPE?\n")
        self.assertEqual(ret, TriggerEdge.FALLING)

    def test_set_trigger_edge(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_trigger_edge(TriggerEdge.RISING)
        self.instr.close()
        self._assert_cmd_no_error([b"TRIGGER:SLOPE POS\n"])

    def test_get_output_load_impedance(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"50\n"
        ret = self.instr.get_output_load_impedance(1)
        self._transport_mock.write.assert_called_once_with(b"OUTPUT1:IMPEDANCE?\n")
        self._transport_mock.write.reset_mock()
        self.assertEqual(ret, 50.0)

        self._transport_mock.read_until.return_value = b"9.9E+37\n"
        ret = self.instr.get_output_load_impedance(2)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"OUTPUT2:IMPEDANCE?\n")
        self.assertEqual(ret, math.inf)

    def test_set_output_load_impedance(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_output_load_impedance(1, 10.0e3)
        self._assert_cmd_no_error([b"OUTPUT1:IMPEDANCE 10000.0\n"])
        self._transport_mock.write.reset_mock()

        self._prep_cmd_no_error()
        self.instr.set_output_load_impedance(2, math.inf)
        self.instr.close()
        self._assert_cmd_no_error([b"OUTPUT2:IMPEDANCE INF\n"])

    def test_get_output_inverted(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"INV\n"
        ret = self.instr.get_output_inverted(1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"OUTPUT1:POLARITY?\n")
        self.assertTrue(ret)

    def test_set_output_inverted(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_output_inverted(1, False)
        self.instr.close()
        self._assert_cmd_no_error([b"OUTPUT1:POLARITY NORM\n"])
        self._transport_mock.write.reset_mock()

    def test_set_display_brightness(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_display_brightness(0.25)
        self.instr.close()
        self._assert_cmd_no_error([b"DISPLAY:BRIGHTNESS 0.250\n"])
        self._transport_mock.write.reset_mock()

    def test_command_error(self):
        self._open_helper()
        self._prep_cmd_with_error()
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_external_reference_enabled(True)
        self.instr.close()
        self._assert_cmd_with_error([b"SOURCE:ROSCILLATOR:SOURCE EXT\n"])

    def test_bad_channel(self):
        self._open_helper()
        with self.assertRaises(ValueError):
            ret = self.instr.get_output_enabled(0)
        with self.assertRaises(ValueError):
            ret = self.instr.get_output_enabled(3)
        self.instr.close()


if __name__ == '__main__':
    unittest.main()
